# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from importlib import import_module
from typing import Any, Callable, Dict, List

from nncf.common.logging import nncf_logger

IMPORTED_DEPENDENCIES: Dict[str, bool] = {}


def skip_if_dependency_unavailable(dependencies: List[str]) -> Callable[[Callable[..., None]], Callable[..., None]]:
    """
    Decorator factory to skip a noreturn function if dependencies are not met.

    :param dependencies: A list of dependencies
    :return: A decorator
    """

    def wrap(func: Callable[..., None]) -> Callable[..., None]:
        def wrapped_f(*args: Any, **kwargs: Any):  # type: ignore
            for libname in dependencies:
                if libname in IMPORTED_DEPENDENCIES:
                    if IMPORTED_DEPENDENCIES[libname]:
                        continue
                    break
                try:
                    _ = import_module(libname)
                    IMPORTED_DEPENDENCIES[libname] = True
                except ImportError as ex:
                    nncf_logger.warning(
                        f"{ex.msg} Please install NNCF package with plots "
                        "extra. Use one of the following commands "
                        '"pip install .[plots]" running from the repository '
                        'root directory or "pip install nncf[plots]"'
                    )
                    IMPORTED_DEPENDENCIES[libname] = False
                    break
            else:
                return func(*args, **kwargs)
            return None

        return wrapped_f

    return wrap



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from typing import Any, Dict, List

from nncf.common.quantization.quantizer_propagation.structs import QuantizerPropagationRule
from nncf.common.quantization.structs import QuantizationPreset
from nncf.common.utils.backend import BackendType
from nncf.quantization.advanced_parameters import AdvancedSmoothQuantParameters
from nncf.quantization.algorithms.bias_correction.algorithm import BiasCorrection
from nncf.quantization.algorithms.channel_alignment.algorithm import ChannelAlignment
from nncf.quantization.algorithms.fast_bias_correction.algorithm import FastBiasCorrection
from nncf.quantization.algorithms.min_max.algorithm import MinMaxQuantization
from nncf.quantization.algorithms.pipeline import Pipeline
from nncf.quantization.algorithms.smooth_quant.algorithm import SmoothQuant
from nncf.quantization.range_estimator import AggregatorType
from nncf.quantization.range_estimator import RangeEstimatorParameters
from nncf.quantization.range_estimator import StatisticsCollectorParameters
from nncf.quantization.range_estimator import StatisticsType

ParamGrid = Dict[str, List[Any]]


def _get_minmax_quantization_param_grid() -> ParamGrid:
    min_param_values = [
        StatisticsCollectorParameters(
            statistics_type=StatisticsType.MIN,
            aggregator_type=AggregatorType.MIN,
        ),
        StatisticsCollectorParameters(
            statistics_type=StatisticsType.QUANTILE,
            aggregator_type=AggregatorType.MEAN,
            quantile_outlier_prob=10e-4,
        ),
        StatisticsCollectorParameters(
            statistics_type=StatisticsType.QUANTILE,
            aggregator_type=AggregatorType.MEAN,
            quantile_outlier_prob=10e-5,
        ),
    ]
    max_param_values = [
        StatisticsCollectorParameters(
            statistics_type=StatisticsType.MAX,
            aggregator_type=AggregatorType.MAX,
        ),
        StatisticsCollectorParameters(
            statistics_type=StatisticsType.QUANTILE,
            aggregator_type=AggregatorType.MEAN,
            quantile_outlier_prob=10e-4,
        ),
        StatisticsCollectorParameters(
            statistics_type=StatisticsType.QUANTILE,
            aggregator_type=AggregatorType.MEAN,
            quantile_outlier_prob=10e-5,
        ),
    ]

    param_grid = {
        "advanced_parameters:quantizer_propagation_rule": [
            QuantizerPropagationRule.MERGE_IF_ALL_BRANCHES_SAME,
        ],
        "preset": [QuantizationPreset.PERFORMANCE, QuantizationPreset.MIXED],
        "advanced_parameters:weights_range_estimator_params": [
            RangeEstimatorParameters(
                min=StatisticsCollectorParameters(statistics_type=StatisticsType.MIN),
                max=StatisticsCollectorParameters(statistics_type=StatisticsType.MAX),
            )
        ],
        "advanced_parameters:activations_range_estimator_params": [
            RangeEstimatorParameters(min=min_v, max=max_v)
            for min_v, max_v in itertools.product(min_param_values, max_param_values)
        ],
    }
    return param_grid


def _get_smooth_quant_param_grid() -> ParamGrid:
    alpha_values = [0.15, 0.25, 0.5, 0.75, 0.95]
    return {
        "advanced_parameters:smooth_quant_alphas": [
            AdvancedSmoothQuantParameters(matmul=alpha_v) for alpha_v in itertools.product(alpha_values)
        ]
    }


def _get_channel_alignment_param_grid() -> ParamGrid:
    return {}


def _get_bias_correction_param_grid() -> ParamGrid:
    return {"fast_bias_correction": [True, False]}


def get_quantization_param_grids(pipeline: Pipeline, backend: BackendType) -> List[ParamGrid]:
    """
    Returns params grid for post-training quantization algorithm.
    """
    algorithm_cls_to_param_grid = {
        SmoothQuant: _get_smooth_quant_param_grid(),
        ChannelAlignment: _get_channel_alignment_param_grid(),
        MinMaxQuantization: _get_minmax_quantization_param_grid(),
        FastBiasCorrection: _get_bias_correction_param_grid(),
        BiasCorrection: _get_bias_correction_param_grid(),
    }

    param_grids = []
    for step in pipeline.pipeline_steps:
        param_grid = {}
        for algorithm in step:
            if backend not in algorithm.available_backends:
                continue
            param_grid.update(algorithm_cls_to_param_grid[algorithm.__class__])
        if param_grid:
            param_grids.append(param_grid)

    return param_grids



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from itertools import islice
from typing import Dict, List, Optional, Tuple

import openvino.runtime as ov

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.data.dataset import Dataset
from nncf.openvino.engine import OVNativeEngine
from nncf.openvino.graph.model_transformer import OVModelTransformer
from nncf.openvino.graph.node_utils import get_parameter_node_name
from nncf.openvino.graph.node_utils import get_result_node_name
from nncf.openvino.graph.transformations.commands import OVStateLessModelExtractionCommand
from nncf.quantization.algorithms.layerwise.iterator import LayerwiseIterator
from nncf.quantization.algorithms.layerwise.scheduler import LayerwiseStep
from nncf.quantization.algorithms.layerwise.scheduler import NodeOutputPort
from nncf.tensor import Tensor


class OVLayerwiseIterator(LayerwiseIterator):
    """
    OVLayerwiseIterator is a class for iterating through layers of an OpenVINO model in a layer-wise manner.
    """

    def __init__(
        self,
        model: ov.Model,
        graph: NNCFGraph,
        schedule: List[LayerwiseStep],
        dataset: Dataset,
        subset_size: int = 100,
        cache: Optional[Dict[NodeOutputPort, List[Tensor]]] = None,
    ):
        """
        :param model: The OpenVINO model to iterate over.
        :param graph: The NNCF graph representation of the model.
        :param schedule: The schedule of steps for layer-wise iteration.
        :param dataset: The dataset to use for model inference.
        :param subset_size: The size of the subset of the dataset to use for each iteration.
        :param cache: The cache for storing tensors.
        """
        self._model = model
        self._graph = graph
        self._schedule = schedule
        self._dataset = dataset
        self._subset_size = subset_size
        self._cache: Dict[NodeOutputPort, List[Tensor]] = {}
        self._cache_lifetime: Dict[NodeOutputPort, int] = {}

        self._step_index = 0
        self._queue: List[Tuple[NNCFNode, Dict[int, List[Tensor]]]] = []

        if cache is not None:
            self.update_cache(cache)

        self._model_transformer = OVModelTransformer(model)
        self._model_input_ids = [NodeOutputPort(node.node_name, 0) for node in graph.get_input_nodes()]
        self._graph_vs_model_inputs_map = {
            input.node.get_friendly_name(): next(iter(input.names)) for input in model.inputs
        }

    def extract_model(self, input_ids: List[NodeOutputPort], output_ids: List[NodeOutputPort]) -> ov.Model:
        """
        Extracts submodel by the specified input & output layers.

        :param input_ids: List of input node IDs.
        :param output_ids: List of output node IDs.
        :return: Extracted OpenVINO model.
        """
        transformation_layout = TransformationLayout()
        model_extraction_command = OVStateLessModelExtractionCommand(input_ids, output_ids)
        transformation_layout.register(model_extraction_command)
        extracted_model = self._model_transformer.transform(transformation_layout)
        return extracted_model

    def create_feed_dicts(self, input_ids: List[NodeOutputPort]) -> List[Dict]:
        """
        Creates a list of dictionaries containing the input data for model execution.

        :param input_ids: List of input node IDs.
        :return: List of dictionaries with the input data.
        """
        subset_size = self._subset_size
        for input_id in input_ids:
            subset_size = min(subset_size, len(self._cache[input_id]))

        feed_dicts = []
        for idx in range(subset_size):
            feed_dict = {}
            for input_id in input_ids:
                if input_id in self._model_input_ids:
                    input_name = self._graph_vs_model_inputs_map[input_id.node_name]
                else:
                    input_name = get_parameter_node_name(input_id.node_name, input_id.output_port)
                feed_dict[input_name] = self._cache[input_id][idx].data
            feed_dicts.append(feed_dict)
        return feed_dicts

    def run_model(
        self, model: ov.Model, feed_dicts: List[Dict], output_ids: List[NodeOutputPort]
    ) -> Dict[NodeOutputPort, List[Tensor]]:
        """
        Runs the submodel with the given input data and collects the output tensors.

        :param model: The OpenVINO submodel to run.
        :param feed_dicts: List of dictionaries containing the input data for the submodel.
        :param output_ids: List of output node IDs to collect outputs from.
        :return: Dictionary of output node IDs to their corresponding tensors.
        """
        outputs = defaultdict(list)
        engine = OVNativeEngine(model)
        for feed_dict in feed_dicts:
            new_output = engine.infer(feed_dict)
            for output_id in output_ids:
                output_name = get_result_node_name(output_id.node_name, output_id.output_port)
                outputs[output_id].append(Tensor(new_output[output_name]))
        return outputs

    def calculate_lifetime(self, input_id: NodeOutputPort) -> int:
        """
        Calculates the lifetime of a tensor in the cache.

        :param input_id: The input node ID.
        :return: The step index at which the tensor is last used.
        """
        for idx in range(len(self._schedule) - 1, self._step_index - 1, -1):
            if input_id in self._schedule[idx].subgraph_inputs:
                return idx
        return -1

    def update_cache(self, outputs: Dict[NodeOutputPort, List[Tensor]]) -> None:
        """
        Updates the cache with new output tensors and removes expired ones.

        :param outputs: Dictionary of output node IDs to their corresponding tensors.
        """
        updated_cache = {}
        updated_cache_lifetime = {}
        for cached_input_id, value in self._cache.items():
            input_step_index = self._cache_lifetime[cached_input_id]
            if input_step_index > self._step_index:
                updated_cache[cached_input_id] = value
                updated_cache_lifetime[cached_input_id] = input_step_index

        for output_id, value in outputs.items():
            output_step_index = self.calculate_lifetime(output_id)
            if output_step_index > self._step_index:
                updated_cache[output_id] = value
                updated_cache_lifetime[output_id] = output_step_index

        self._cache = updated_cache
        self._cache_lifetime = updated_cache_lifetime

    def collect_output_tensors(self, step: LayerwiseStep) -> Dict[NodeOutputPort, List[Tensor]]:
        """
        Collects the output tensors for a given step.

        :param step: The current layer-wise step.
        :return: Dictionary of output node IDs to their corresponding tensors.
        """
        outputs = {}
        for output_id in step.subgraph_outputs:
            if output_id not in self._cache:
                break
            outputs[output_id] = self._cache[output_id]
        else:
            return outputs

        subgraph_model_input_ids = []
        for input_id in step.subgraph_inputs:
            if input_id in self._cache:
                continue
            if input_id in self._model_input_ids:
                subgraph_model_input_ids.append(input_id)
            else:
                raise RuntimeError(
                    f"{input.node_name}:{input.output_port} is not found in the input cache and is not a model input"
                )

        if subgraph_model_input_ids:
            subgraph_inputs = self._model_input_ids
            subgraph_outputs = [output_id for output_id in step.subgraph_outputs]
            for step in self._schedule[self._step_index + 1 :]:
                for input_id in step.subgraph_inputs:
                    if (
                        input_id in self._model_input_ids
                        and input_id not in self._cache
                        and input_id not in subgraph_outputs
                    ):
                        subgraph_outputs.append(input_id)
            feed_dicts = islice(self._dataset.get_inference_data(), self._subset_size)
        else:
            subgraph_inputs = step.subgraph_inputs
            subgraph_outputs = step.subgraph_outputs
            feed_dicts = self.create_feed_dicts(step.subgraph_inputs)
        extracted_model = self.extract_model(subgraph_inputs, subgraph_outputs)
        return self.run_model(extracted_model, feed_dicts, subgraph_outputs)

    def __next__(self) -> Tuple[NNCFNode, Dict[int, List[Tensor]]]:
        """
        Advances to the next layer in the iteration.

        :return: A tuple containing the target node and a dictionary of input port IDs to tensors.
        :raises StopIteration: If the iteration has reached the end of the schedule.
        """
        if not self._queue:
            if self._step_index >= len(self._schedule):
                raise StopIteration

            step = self._schedule[self._step_index]
            outputs = self.collect_output_tensors(step)

            self.update_cache(outputs)

            for target_node, target_inputs in step.target_node_map.items():
                inputs = {input_port_id: outputs[output_id] for input_port_id, output_id in target_inputs.items()}
                self._queue.append((target_node, inputs))
            self._step_index += 1

        return self._queue.pop(0)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional

import tensorflow as tf

import nncf
from nncf.common.initialization.dataloader import NNCFDataLoader
from nncf.common.quantization.structs import QuantizationPreset
from nncf.config import NNCFConfig
from nncf.config.structures import BNAdaptationInitArgs
from nncf.config.structures import QuantizationRangeInitArgs
from nncf.data import Dataset
from nncf.data.dataset import DataProvider
from nncf.parameters import ModelType
from nncf.parameters import QuantizationMode
from nncf.parameters import TargetDevice
from nncf.quantization.advanced_parameters import AdvancedQuantizationParameters
from nncf.quantization.advanced_parameters import apply_advanced_parameters_to_config
from nncf.scopes import IgnoredScope
from nncf.scopes import convert_ignored_scope_to_list
from nncf.tensorflow.helpers.model_creation import create_compressed_model_impl

DEFAULT_RANGE_TYPE = "mean_min_max"


# TODO(alexsu52): It is a workaround and should be removed.
class CalibrationDataLoader(NNCFDataLoader):
    """
    This class wraps the nncf.Dataset.

    This is required for proper initialization of certain compression algorithms.
    """

    def __init__(self, dataset: Dataset):
        self._dataset = dataset

    @property
    def batch_size(self) -> int:
        data_source = getattr(self._dataset, "_data_source")

        if not hasattr(data_source, "_batch_size"):
            return 1
        batch_size = getattr(data_source, "_batch_size")
        try:
            if isinstance(batch_size, tf.Tensor):
                batch_size = batch_size.numpy()
            batch_size = int(batch_size)
        except:  # noqa: E722
            batch_size = 1
        return batch_size

    def __iter__(self):
        def transform_fn(data_item):
            return data_item, None

        return iter(DataProvider(self._dataset.get_inference_data(), transform_fn))


def _get_default_quantization_config(preset: QuantizationPreset, subset_size: int) -> Dict[str, Any]:
    """
    Returns the default quantization config

    :param preset: A preset that controls the quantization mode
        (symmetric and asymmetric). It can take the following values:
        - `performance`: Symmetric quantization of weights and activations.
        - `mixed`: Symmetric quantization of weights and asymmetric
          quantization of activations.
    :param subset_size: Size of a subset to calculate activations
        statistics used for quantization.
    :return: The default quantization config.
    """
    return {
        "algorithm": "quantization",
        "preset": preset.value,
        "initializer": {
            "range": {"num_init_samples": subset_size, "type": DEFAULT_RANGE_TYPE},
            "batchnorm_adaptation": {"num_bn_adaptation_samples": subset_size},
        },
        "overflow_fix": "first_layer_only",
    }


def _create_nncf_config(
    preset: QuantizationPreset,
    target_device: TargetDevice,
    subset_size: int,
    ignored_scope: Optional[IgnoredScope],
    advanced_parameters: Optional[AdvancedQuantizationParameters],
) -> NNCFConfig:
    """
    Creates the NNCFConfig for the quantization algorithm.

    :param preset: A preset that controls the quantization mode
        (symmetric and asymmetric). It can take the following values:
        - `performance`: Symmetric quantization of weights and activations.
        - `mixed`: Symmetric quantization of weights and asymmetric
          quantization of activations.
    :param target_device: A target device the specificity of which will be taken
        into account while compressing in order to obtain the best performance
        for this type of device.
    :param subset_size: Size of a subset to calculate activations
        statistics used for quantization.
    :param ignored_scope:  An ignored scope that defined the list of model control
        flow graph nodes to be ignored during quantization.
    :param advanced_parameters: Advanced quantization parameters for
        fine-tuning the quantization algorithm.
    :return: NNCFConfig for the quantization algorithm.
    """
    compression_config = _get_default_quantization_config(preset, subset_size)

    if ignored_scope is not None:
        _ignored_scope = convert_ignored_scope_to_list(ignored_scope)
        if "ignored_scopes" in compression_config:
            compression_config["ignored_scopes"].extend(_ignored_scope)
        else:
            compression_config["ignored_scopes"] = _ignored_scope
        compression_config["validate_scopes"] = ignored_scope.validate

    if advanced_parameters is not None:
        compression_config = apply_advanced_parameters_to_config(compression_config, advanced_parameters)

    return NNCFConfig({"target_device": target_device.value, "compression": compression_config})


def quantize_impl(
    model: tf.Module,
    calibration_dataset: Dataset,
    mode: Optional[QuantizationMode] = None,
    preset: Optional[QuantizationPreset] = None,
    target_device: TargetDevice = TargetDevice.ANY,
    subset_size: int = 300,
    fast_bias_correction: bool = True,
    model_type: Optional[ModelType] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    advanced_parameters: Optional[AdvancedQuantizationParameters] = None,
) -> tf.Module:
    """
    Implementation of the `quantize()` method for the TensorFlow backend.
    """
    if model_type is not None:
        raise ValueError(f"model_type={model_type} is not supported")
    if fast_bias_correction is False:
        raise ValueError(f"fast_bias_correction={fast_bias_correction} is not supported")
    if ignored_scope is not None and ignored_scope.types:
        raise nncf.InternalError(
            "Quantization algorithm form the TensorFlow backend "
            "does not support operation types in the ignored "
            "scopes yet"
        )
    if target_device == TargetDevice.CPU_SPR:
        raise nncf.InternalError("target_device == CPU_SPR is not supported.")

    if mode is not None:
        raise ValueError(f"mode={mode} is not supported")

    if preset is None:
        preset = QuantizationPreset.PERFORMANCE

    nncf_config = _create_nncf_config(preset, target_device, subset_size, ignored_scope, advanced_parameters)

    calibration_data_loader = CalibrationDataLoader(calibration_dataset)
    nncf_config.register_extra_structs(
        [
            QuantizationRangeInitArgs(data_loader=calibration_data_loader),
            BNAdaptationInitArgs(data_loader=calibration_data_loader),
        ]
    )

    compression_ctrl, compressed_model = create_compressed_model_impl(model=model, config=nncf_config)

    # NOTE: We set the config here to properly save/load the quantized model during training into tf.train.Checkpoint.
    # You can obtain that config via the nncf.tensorflow.get_config() method and save/load it to/from
    # tf.train.Checkpoint using the nncf.tensorflow.ConfigState class.
    config = compression_ctrl.get_compression_state()["builder_state"]
    setattr(compressed_model, "_nncf_config", config)

    return compressed_model



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict

import tensorflow as tf

from nncf.api.compression import CompressionLoss


class SparseLoss(CompressionLoss):
    def __init__(self, target_ops, target=1.0, p=0.05):
        super().__init__()
        self._target_ops = target_ops
        self.target = tf.Variable(target, trainable=False)
        self.p = p
        self.disabled = tf.Variable(False, trainable=False)

    def disable(self):
        tf.cond(tf.cast(self.disabled, tf.bool), lambda: None, self._disable)

    def _disable(self):
        self.disabled.assign(True)

        for op, op_weights in self._target_ops:
            op.freeze(op_weights)

    def calculate(self, *args, **kwargs):
        return tf.cond(tf.cast(self.disabled, tf.bool), lambda: tf.constant(0.0), self._calculate)

    def _calculate(self):
        params = tf.constant(0)
        loss = tf.constant(0.0)
        for op, op_weights in self._target_ops:
            tf.debugging.assert_equal(
                op.get_trainable_weight(op_weights),
                tf.constant(True),
                "Invalid state of SparseLoss and SparsifiedWeight: mask is frozen for enabled loss",
            )
            mask = op.get_mask(op_weights)
            params = params + tf.size(mask)
            loss = loss + op.loss(op_weights)

        params = tf.cast(params, tf.float32)
        return tf.reshape(tf.math.pow(((loss / params - self.target) / self.p), 2), shape=[])

    @property
    def target_sparsity_rate(self):
        eager_target = tf.keras.backend.eval(self.target)
        rate = 1.0 - eager_target
        if rate < 0 or rate > 1:
            raise ValueError("Target is not within range [0, 1]")
        return rate

    def set_target_sparsity_loss(self, sparsity_level):
        self.target.assign(1 - sparsity_level)

    def load_state(self, state: Dict[str, Any]) -> None:
        self.target.assign(state["target"])
        self.disabled.assign(state["disabled"])
        self.p = state["p"]

    def get_state(self) -> Dict[str, Any]:
        return {
            "target": float(tf.keras.backend.eval(self.target)),
            "disabled": bool(tf.keras.backend.eval(tf.cast(self.disabled, tf.bool))),
            "p": self.p,
        }



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import deque
from typing import List

import openvino.runtime as ov

from nncf.common.factory import ModelTransformerFactory
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.openvino.graph.metatypes.groups import FAKE_QUANTIZE_OPERATIONS
from nncf.openvino.graph.metatypes.openvino_metatypes import OVReadValueMetatype
from nncf.openvino.graph.transformations.command_creation import OVCommandCreator


def remove_fq_from_inputs(model: ov.Model, graph: NNCFGraph) -> ov.Model:
    """
    This method removes the activation Fake Quantize nodes from the model.
    It's needed for the further bias shift calculation that relates on quantized weights.

    :param model: ov.Model instance.
    :param graph: NNCFGraph instance.
    :return: ov.Model instance without activation Fake Quantize nodes.
    """
    transformation_layout = TransformationLayout()
    model_transformer = ModelTransformerFactory.create(model)

    seen_nodes = []
    nodes_queue = deque(graph.get_input_nodes())
    while nodes_queue:
        current_node = nodes_queue.popleft()
        current_node_name = current_node.node_name

        if current_node_name in seen_nodes:
            continue

        seen_nodes.append(current_node_name)
        if current_node.metatype in FAKE_QUANTIZE_OPERATIONS:
            command = OVCommandCreator.create_command_to_remove_quantizer(current_node)
            transformation_layout.register(command)
        nodes_queue.extend(graph.get_next_nodes(current_node))

    return model_transformer.transform(transformation_layout)


def get_start_nodes_for_activation_path_tracing(nncf_graph: NNCFGraph) -> List[NNCFNode]:
    """
    Get a list of NNCFNodes to use as start nodes for activation path tracing.

    :param nncf_graph: NNCFGraph to work with.
    :return: Target NNCFGraph input nodes.
    """
    return nncf_graph.get_input_nodes() + nncf_graph.get_nodes_by_metatypes([OVReadValueMetatype])


def remove_friendly_name_duplicates(model: ov.Model) -> ov.Model:
    """
    Removes diplicates of node names (friendly_name attribute) in the model.

    :param model: ov.Model instance to update.
    :return: Updated ov.Model without duplicated names.
    """
    rt_info_path = ["nncf", "friendly_names_were_updated"]
    friendly_names_flag = "True"
    if model.has_rt_info(rt_info_path) and model.get_rt_info(rt_info_path).value == friendly_names_flag:
        return model

    existing_names = set()
    for op in model.get_ops():
        friendly_name = op.get_friendly_name()
        if friendly_name in existing_names:
            friendly_name = friendly_name + "0"
            op.set_friendly_name(friendly_name)
        existing_names.add(friendly_name)
    model.set_rt_info(friendly_names_flag, rt_info_path)
    return model


def model_has_state(model: ov.Model) -> bool:
    """
    Returns True if model has state else False

    :param model: OpenVINO model
    :return: True if model has state else False
    """
    return len(model.get_sinks()) > 0


def copy_rt_info(model_source: ov.Model, model_dest: ov.Model, path: List[str]) -> None:
    """
    Checks and copies the rt_info from the source to destination model.

    :param model_source: ov.Model instance to copy rt_info from.
    :param model_dest: ov.Model instance to copy rt_info to.
    :param path: Path to rt_info.
    """
    if model_source.has_rt_info(path):
        source_rt_info = model_source.get_rt_info(path)
        model_dest.set_rt_info(source_rt_info, path)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
from copy import deepcopy
from typing import List

import torch
import torch.distributed as dist

import nncf
from nncf import NNCFConfig
from nncf.api.compression import CompressionStage
from nncf.common.accuracy_aware_training.training_loop import ADAPTIVE_COMPRESSION_CONTROLLERS
from nncf.common.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.logging import nncf_logger
from nncf.common.scopes import matches_any
from nncf.common.sparsity.statistics import MovementSparsityStatistics
from nncf.common.statistics import NNCFStatistics
from nncf.config.extractors import extract_algo_specific_config
from nncf.experimental.torch.sparsity.movement.layers import MovementSparsifier
from nncf.experimental.torch.sparsity.movement.layers import SparseConfig
from nncf.experimental.torch.sparsity.movement.layers import SparseConfigByScope
from nncf.experimental.torch.sparsity.movement.layers import SparseStructure
from nncf.experimental.torch.sparsity.movement.loss import ImportanceLoss
from nncf.experimental.torch.sparsity.movement.scheduler import MovementPolynomialThresholdScheduler
from nncf.experimental.torch.sparsity.movement.scheduler import MovementSchedulerParams
from nncf.experimental.torch.sparsity.movement.structured_mask_handler import StructuredMaskHandler
from nncf.torch.algo_selector import PT_COMPRESSION_ALGORITHMS
from nncf.torch.compression_method_api import PTCompressionAlgorithmController
from nncf.torch.graph.transformations.commands import PTInsertionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.graph.transformations.commands import TransformationPriority
from nncf.torch.layers import NNCFLinear
from nncf.torch.module_operations import UpdateWeightAndBias
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.sparsity.base_algo import BaseSparsityAlgoBuilder
from nncf.torch.sparsity.base_algo import BaseSparsityAlgoController
from nncf.torch.sparsity.base_algo import SparseModuleInfo
from nncf.torch.sparsity.collector import PTSparseModelStatisticsCollector
from nncf.torch.utils import get_model_device

SUPPORTED_NNCF_MODULES = [NNCFLinear]


@PT_COMPRESSION_ALGORITHMS.register("movement_sparsity")
class MovementSparsityBuilder(BaseSparsityAlgoBuilder):
    def __init__(self, config, should_init: bool = True):
        super().__init__(config, should_init)
        configs = self._algo_config.get("sparse_structure_by_scopes", [])
        self._sparse_configs_by_scopes = [SparseConfigByScope.from_config(c) for c in configs]

    def create_weight_sparsifying_operation(
        self, target_module_node: NNCFNode, compression_lr_multiplier: float
    ) -> MovementSparsifier:
        sparse_cfg = SparseConfig(SparseStructure.FINE)
        node_name = target_module_node.node_name
        matched_scopes = []
        for configs_per_scopes in self._sparse_configs_by_scopes:
            target_scopes = configs_per_scopes.target_scopes
            if matches_any(node_name, target_scopes):
                sparse_cfg = configs_per_scopes.sparse_config
                matched_scopes.append(target_scopes)
        if len(matched_scopes) >= 2:
            raise nncf.InternalError(f'"{node_name}" is matched by multiple items in `sparse_structure_by_scopes`.')

        return MovementSparsifier(
            target_module_node,
            sparse_cfg=sparse_cfg,
            frozen=False,
            compression_lr_multiplier=compression_lr_multiplier,
            layerwise_loss_lambda=0.5,
        )

    def _sparsify_weights(self, target_model: NNCFNetwork) -> List[PTInsertionCommand]:
        device = get_model_device(target_model)
        sparsified_module_nodes = target_model.nncf.get_weighted_original_graph_nodes(
            nncf_module_names=[m.__name__ for m in SUPPORTED_NNCF_MODULES]
        )
        insertion_commands = []
        for module_node in sparsified_module_nodes:
            node_name = module_node.node_name

            if not self._should_consider_scope(node_name):
                nncf_logger.info(f"Ignored adding weight sparsifier in scope: {node_name}")
                continue

            nncf_logger.debug("Adding weight sparsifier in scope: {node_name}")
            compression_lr_multiplier = self.config.get_redefinable_global_param_value_for_algo(
                "compression_lr_multiplier", self.name
            )
            sparsifying_operation = self.create_weight_sparsifying_operation(module_node, compression_lr_multiplier)
            hook = UpdateWeightAndBias(sparsifying_operation).to(device)
            insertion_commands.append(
                PTInsertionCommand(
                    PTTargetPoint(TargetType.PRE_LAYER_OPERATION, target_node_name=node_name),
                    hook,
                    TransformationPriority.SPARSIFICATION_PRIORITY,
                )
            )
            sparsified_module = target_model.nncf.get_containing_module(node_name)
            self._sparsified_module_info.append(SparseModuleInfo(node_name, sparsified_module, sparsifying_operation))

        if not insertion_commands:
            raise nncf.InternalError("No sparsifiable layer found for movement sparsity algorithm.")
        return insertion_commands

    def _build_controller(self, model: NNCFNetwork) -> PTCompressionAlgorithmController:
        return MovementSparsityController(model, self._sparsified_module_info, self.config)


MODEL_FAMILIES = ["bert", "wav2vec2", "swin", "mobilebert", "distilbert", "clip", "vit"]


def is_supported_model_family(model: NNCFNetwork) -> None:
    """
    Checks whether the model family is supported by movement sparsity to conduct structured masking.

    :param model: The compressed model wrapped by NNCF.
    """
    model_pymodules = inspect.getmodule(model).__name__.split(".")
    is_supported = False
    if len(model_pymodules) >= 3 and model_pymodules[:2] == ["transformers", "models"]:
        # the case of input model defined by HuggingFace's transformers
        model_family = model_pymodules[2]
        is_supported = model_family in MODEL_FAMILIES
    return is_supported


@ADAPTIVE_COMPRESSION_CONTROLLERS.register("pt_movement_sparsity")
class MovementSparsityController(BaseSparsityAlgoController):
    def __init__(self, target_model: NNCFNetwork, sparsified_module_info: List[SparseModuleInfo], config: NNCFConfig):
        super().__init__(target_model, sparsified_module_info)
        algo_config = extract_algo_specific_config(config, "movement_sparsity")
        sparsify_operations = [m.operand for m in self.sparsified_module_info]
        params = deepcopy(algo_config.get("params", {}))
        self._distributed = False
        self._scheduler_params = MovementSchedulerParams.from_dict(params)
        self._scheduler = MovementPolynomialThresholdScheduler(self, self._scheduler_params)
        self._loss = ImportanceLoss(sparsify_operations)
        self._config = config

        if self._scheduler.enable_structured_masking:
            if not is_supported_model_family(self.model):
                raise nncf.UnsupportedModelError(
                    "You set `enable_structured_masking=True`, but no supported model is detected. "
                    f"Supported model families: {MODEL_FAMILIES}."
                )
            self._structured_mask_handler = StructuredMaskHandler(self.model, self.sparsified_module_info)

    @property
    def compression_rate(self) -> float:
        return self.statistics().movement_sparsity.model_statistics.sparsity_level

    def reset_independent_structured_mask(self):
        """
        Asks the structured mask handler to gather independent masks in the model.
        """
        assert self._scheduler.enable_structured_masking
        self._structured_mask_handler.update_independent_structured_mask()

    def resolve_structured_mask(self):
        """
        Asks the structured mask handler to resolve dependent masks in the model.
        """
        assert self._scheduler.enable_structured_masking
        self._structured_mask_handler.resolve_dependent_structured_mask()

    def populate_structured_mask(self):
        """
        Asks the structured mask handler to update structured binary masks in model operands.
        """
        assert self._scheduler.enable_structured_masking
        self._structured_mask_handler.populate_dependent_structured_mask_to_operand()
        self._structured_mask_handler.report_structured_sparsity(self._config.get("log_dir", "."))

    def compression_stage(self) -> CompressionStage:
        if self.scheduler.current_epoch < self._scheduler_params.warmup_start_epoch:
            return CompressionStage.UNCOMPRESSED
        if self.scheduler.current_epoch >= self._scheduler_params.warmup_end_epoch:
            return CompressionStage.FULLY_COMPRESSED
        return CompressionStage.PARTIALLY_COMPRESSED

    def distributed(self):
        if not dist.is_initialized():
            raise KeyError(
                "Could not set distributed mode for the compression algorithm "
                "because the default process group has not been initialized."
            )

        if next(self._model.parameters()).is_cuda:
            state = torch.cuda.get_rng_state()
            if dist.get_backend() == dist.Backend.NCCL:
                state = state.cuda()
            torch.distributed.broadcast(state, src=0)
            torch.cuda.set_rng_state(state.cpu())
        else:
            state = torch.get_rng_state()
            torch.distributed.broadcast(state, src=0)
            torch.set_rng_state(state)

        self._distributed = True

    def freeze(self):
        self._loss.disable()
        for minfo in self.sparsified_module_info:
            minfo.operand.requires_grad_(False)

    def statistics(self, quickly_collected_only=False) -> NNCFStatistics:
        collector = PTSparseModelStatisticsCollector(self.model, self.sparsified_module_info, supports_sparse_bias=True)
        model_statistics = collector.collect()

        stats = MovementSparsityStatistics(
            model_statistics,
            self.scheduler.current_importance_threshold,
            self.scheduler.current_importance_regularization_factor,
        )

        nncf_stats = NNCFStatistics()
        nncf_stats.register("movement_sparsity", stats)
        return nncf_stats



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple, Union

from nncf.common.graph.utils import get_reduction_axes
from nncf.common.initialization.dataloader import NNCFDataLoader
from nncf.common.quantization.structs import QuantizationScheme
from nncf.common.quantization.structs import QuantizerGroup
from nncf.common.tensor_statistics.collectors import ReductionAxes
from nncf.config.schemata.defaults import NUM_INIT_SAMPLES
from nncf.experimental.common.tensor_statistics.collectors import AggregationAxes


class RangeInitConfig:
    """
    The `RangeInitConfig` class representing the quantization range initialization
    parameters.
    """

    def __init__(
        self,
        init_type: str,
        num_init_samples: int,
        init_type_specific_params: Optional[Dict[str, int]] = None,
    ):
        """
        Initializes the quantization range initialization parameters.

        :param init_type: Type of the initializer - determines which tensor
            statistics will be used to initialize quantization ranges.
        :param num_init_samples: The number of samples from the dataset to consume
            as sample model inputs to compute quantization ranges.
        :param init_type_specific_params: additional parameters specific to the type
            of the initializer
        """
        self.init_type = init_type
        self.num_init_samples = num_init_samples
        self.init_type_specific_params = init_type_specific_params
        if self.init_type_specific_params is None:
            self.init_type_specific_params = {}

    def __eq__(self, other: object) -> bool:
        return self.__dict__ == other.__dict__

    @classmethod
    def from_dict(cls, dct: Dict[str, Any]) -> RangeInitConfig:
        num_init_samples = dct.get("num_init_samples", NUM_INIT_SAMPLES)
        if num_init_samples < 0:
            raise ValueError("Number of initialization samples must be >= 0")
        return cls(dct.get("type", "mixed_min_max"), num_init_samples, dct.get("params"))


class PerLayerRangeInitConfig(RangeInitConfig):
    """
    The `PerLayerRangeInitConfig` class representing the quantization range
    initialization parameters for layers which are specified using the target
    and ignored scopes and the target group of quantizers.
    """

    def __init__(
        self,
        range_init_config: RangeInitConfig,
        target_scopes: Optional[List[str]],
        ignored_scopes: Optional[List[str]],
        target_quantizer_group: QuantizerGroup = None,
    ):
        """
        Initializes the quantization range initialization parameters.

        :param range_init_config: The quantization range initialization parameters.
        :param target_scopes: A list of model control flow graph node scopes
            to be considered for this operation - functions as a 'denylist'
        :param ignored_scopes: A list of model control flow graph node scopes
            to be ignored for this operation - functions as an 'allowlist'
        :param target_quantizer_group: The target group of quantizers for which
            specified type of range initialization will be applied. It can be
            quantizers group for activations or weights.
        """

        super().__init__(
            range_init_config.init_type, range_init_config.num_init_samples, range_init_config.init_type_specific_params
        )
        if target_scopes is None and ignored_scopes is None:
            raise ValueError(
                "At least one of the (target_scopes, ignored_scopes) should be specified"
                " for a per-layer range init config!"
            )
        self.target_scopes = target_scopes
        self.ignored_scopes = ignored_scopes
        self.target_group = target_quantizer_group

    @classmethod
    def from_dict(cls, dct: Dict[str, Any]) -> PerLayerRangeInitConfig:
        base_config = RangeInitConfig.from_dict(dct)

        def get_list(dct: Dict[str, Any], attr_name: str) -> Optional[List[str]]:
            str_or_list = dct.get(attr_name)
            if str_or_list is None:
                return None
            if isinstance(str_or_list, str):
                retval_list = [str_or_list]
            else:
                retval_list = str_or_list
            return retval_list

        target_scopes, ignored_scopes = get_list(dct, "target_scopes"), get_list(dct, "ignored_scopes")

        target_group_str = dct.get("target_quantizer_group")
        target_group = None
        if target_group_str is not None:
            target_group = QuantizerGroup(target_group_str)

        return cls(base_config, target_scopes, ignored_scopes, target_group)


class RangeInitParams:
    """
    The `RangeInitParams` class representing the initialization dataset and the
    quantization range initialization parameters for all model layers.
    """

    def __init__(
        self,
        init_range_data_loader: NNCFDataLoader,
        device: str,
        global_init_config: Optional[RangeInitConfig],
        per_layer_range_init_configs: List[PerLayerRangeInitConfig],
    ):
        """

        :param init_range_data_loader: Provides an iterable over the given dataset.
        :param device: Device to perform initialization. If `device` is `None`
            then the device of the model parameters will be used.
        :param global_init_config: The quantization range initialization parameters
            for a model
        :param per_layer_range_init_configs: The List of the quantization range
            initialization parameters for model layers
        """
        self.init_range_data_loader = init_range_data_loader
        self.device = device
        self.global_init_config = global_init_config
        self.per_layer_range_init_configs = per_layer_range_init_configs


class RangeInitCollectorParams:
    """
    Defines low-level parameters that are used to instantiate statistic collectors.
    """

    def __init__(self, is_weights: bool, scheme: QuantizationScheme, per_channel: bool):
        """
        Initializes Range Initialization Collector Parameters.

        :param is_weights: Boolean that defines tensor type. True for Weights, False for Activations.
        :param scheme: Quantization scheme: symmetric or asymmetric.
        :param per_channel: Quantization granularity.
        """
        self._is_weights = is_weights
        self._scheme = scheme
        self._is_per_channel = per_channel

    @property
    def is_weights(self) -> bool:
        """
        Returns boolean that defines tensor type.
        True for Weights, False for Activations.
        """
        return self._is_weights

    @property
    def scheme(self) -> QuantizationScheme:
        """
        Returns quantization scheme: symmetric or asymmetric.
        """
        return self._scheme

    @property
    def is_per_channel(self) -> bool:
        """
        Returns quantization granularity.
        """
        return self._is_per_channel

    def use_per_sample_stats(self, per_sample_stats: bool) -> bool:
        """
        For activations, if per_sample_stats is True, statistics will be collected per-sample.
        For weights statistics are always collected per-batch.

        :param per_sample_stats: Defined by certain collector design.
        :return: A boolean that defines whether to collect statistics per-sample or per-batch.
        """
        return per_sample_stats and (not self._is_weights)

    @property
    def use_abs_max(self) -> bool:
        """Applies abs(max) for symmetric quantization."""
        return self._scheme == QuantizationScheme.SYMMETRIC

    @property
    def use_means_of_mins(self) -> bool:
        return not self._is_weights and not self._is_per_channel and self._scheme == "asymmetric"

    @property
    def use_means_of_maxs(self) -> bool:
        return not self._is_weights and not self._is_per_channel

    def _get_reduction_axes(
        self,
        shape_to_reduce: Union[Tuple[int, ...], List[int]],
        quantization_axes: Union[Tuple[int, ...], List[int]],
        aggregation_axes: Union[Tuple[int, ...], List[int]],
    ) -> Tuple[int, ...]:
        """
        Returns axes for a reducer regarding aggregation axes. As aggregator takes axes counting from stacked tensors,
        from these axes only tensor related axes should be used for reducer.

        :param shape_to_reduce: Shape of a reduced tensor.
        :param quantization_axes: Axes of quantization.
        :param aggregation_axes: Axes of aggregator which is applied onto reduced tensor.
        :return: Axes for reducer.
        """
        axes_to_keep = set(el - 1 for el in aggregation_axes if el != 0)
        axes_to_keep.update(quantization_axes)
        return get_reduction_axes(list(axes_to_keep), shape_to_reduce)

    def _get_aggregation_axes(self, batchwise_statistics: bool) -> Tuple[int, ...]:
        """
        Returns axes for aggregator.

        :param batchwise_statistics: Determines whether quantizer statistics should be calculated
            for each item of the batch or for the entire batch.
        :return Tuple[int]: Aggregation axes.
        """
        return (0, 1) if batchwise_statistics else (0,)

    def get_reduction_aggregation_axes(
        self,
        shape_to_reduce: Union[Tuple[int, ...], List[int]],
        quantization_axes: Union[Tuple[int, ...], List[int]],
        batchwise_statistics: bool,
    ) -> Tuple[ReductionAxes, AggregationAxes]:
        """
        Calculates the reduction axes, aggregation axes for the tensor.

        :param shape_to_reduce: Shape of the tensor.
        :param quantization_axes: Quantization axes if per-channel quantization.
        :param batchwise_statistics: Determines whether quantizer statistics should be calculated
            for each item of the batch or for the entire batch.
        :return: Reduction axes and aggregation axes.
        """
        aggregation_axes = self._get_aggregation_axes(batchwise_statistics)
        reduction_axes = self._get_reduction_axes(shape_to_reduce, quantization_axes, aggregation_axes)
        return reduction_axes, aggregation_axes



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from collections import Counter
from typing import List, Tuple

from nncf.common.collector import StatisticsCollector
from nncf.common.quantization.statistics import QuantizationStatistics
from nncf.common.quantization.statistics import QuantizersCounter


class QuantizerDescription:
    """
    Contains information about the quantizer.
    """

    def __init__(
        self,
        num_bits: int,
        is_per_channel: bool,
        is_signed: bool,
        is_symmetric: bool,
        is_weight_quantizer: bool,
        is_enabled: bool,
    ):
        """
        Initializes the description of the quantizer.

        :param num_bits: Bitwidth of the quantization.
        :param is_per_channel: `True` for per-channel quantization, `False` for per-tensor.
        :param is_signed: `True` for signed quantization, `False` for unsigned.
        :param is_symmetric: `True` for symmetric quantizer, `False` for asymmetric.
        :param is_weight_quantizer: `True` for weight quantizer, `False` for non-weight.
        :param is_enabled: `True` for enabled quantizer, `False` for disabled.
        """
        self._num_bits = num_bits
        self._is_per_channel = is_per_channel
        self._is_signed = is_signed
        self._is_symmetric = is_symmetric
        self._is_weight_quantizer = is_weight_quantizer
        self._is_enabled = is_enabled

    @property
    def num_bits(self) -> int:
        return self._num_bits

    @property
    def is_per_channel(self) -> bool:
        return self._is_per_channel

    @property
    def is_signed(self) -> bool:
        return self._is_signed

    @property
    def is_symmetric(self) -> bool:
        return self._is_symmetric

    @property
    def is_weight_quantizer(self) -> bool:
        return self._is_weight_quantizer

    @property
    def is_enabled(self) -> bool:
        return self._is_enabled


class QuantizationStatisticsCollector(StatisticsCollector):
    """
    Base class for the quantization statistics collector.
    """

    @abstractmethod
    def _collect_quantizers_descriptions(self) -> List[QuantizerDescription]:
        """
        Collects descriptions of the quantizers.

        :return: Descriptions of the quantizers.
        """

    @abstractmethod
    def _get_potential_quantizers_num(self) -> Tuple[int, int]:
        """
        Returns a potential number of quantizers for weights and activations.

        :return: A tuple (wq_potential_num, aq_potential_num) where
            - `wq_potential_num` is a potential number of quantizers for weights.
            - `aq_potential_num` is a potential number of quantizers for activations.
        """

    def collect(self) -> QuantizationStatistics:
        """
        Collects statistics of the quantization algorithm.

        :return: A statistics of the quantization algorithm.
        """
        quantizers_descriptions = self._collect_quantizers_descriptions()

        wq_counter = QuantizersCounter()
        aq_counter = QuantizersCounter()
        wq_bitwidths = []
        aq_bitwidths = []
        num_enabled_quantizers = 0

        wq_counter.potential_count, aq_counter.potential_count = self._get_potential_quantizers_num()

        for q in quantizers_descriptions:
            if q.is_weight_quantizer:
                counter = wq_counter
                wq_bitwidths.append(q.num_bits)
            else:
                counter = aq_counter
                aq_bitwidths.append(q.num_bits)

            if q.is_per_channel:
                counter.num_per_channel += 1
            else:
                counter.num_per_tensor += 1

            if q.is_signed:
                counter.num_signed += 1
            else:
                counter.num_unsigned += 1

            if q.is_symmetric:
                counter.num_symmetric += 1
            else:
                counter.num_asymmetric += 1

            if q.is_enabled:
                num_enabled_quantizers += 1

            counter.total_count += 1

        num_wq_per_bitwidth = dict(Counter(wq_bitwidths))
        num_aq_per_bitwidth = dict(Counter(aq_bitwidths))

        total_count = wq_counter.total_count + aq_counter.total_count
        ratio_of_enabled_quantizations = 100 * (num_enabled_quantizers / max(total_count, 1))

        return QuantizationStatistics(
            wq_counter, aq_counter, num_wq_per_bitwidth, num_aq_per_bitwidth, ratio_of_enabled_quantizations
        )



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List, Tuple

import numpy as np
import openvino.runtime as ov

import nncf
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.tensor_statistics.statistic_point import StatisticPoint
from nncf.experimental.common.tensor_statistics.collectors import MaxAggregator
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.openvino.graph.layout import OVLayoutElem
from nncf.openvino.graph.layout import get_linear_weights_layout_from_node
from nncf.openvino.graph.metatypes.groups import QUANTIZE_AGNOSTIC_OPERATIONS
from nncf.openvino.graph.metatypes.openvino_metatypes import OVConvolutionMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVMatMulMetatype
from nncf.openvino.graph.node_utils import get_weight_value
from nncf.openvino.graph.transformations.command_creation import OVCommandCreator
from nncf.openvino.graph.transformations.commands import OVMultiplyInsertionCommand
from nncf.openvino.graph.transformations.commands import OVTargetPoint
from nncf.openvino.graph.transformations.commands import OVWeightUpdateCommand
from nncf.openvino.statistics.collectors import OVAbsMaxReducer
from nncf.quantization.algorithms.smooth_quant.backend import SmoothQuantAlgoBackend
from nncf.tensor import Tensor

OV_PRE_LAYER_TARGET_TYPE = TargetType.PRE_LAYER_OPERATION


class OVSmoothQuantAlgoBackend(SmoothQuantAlgoBackend):
    @property
    def convolution_metatypes(self) -> List[OperatorMetatype]:
        return [OVConvolutionMetatype]

    @property
    def matmul_metatypes(self) -> List[OperatorMetatype]:
        return [OVMatMulMetatype]

    @property
    def quantize_agnostic_metatypes(self) -> List[OperatorMetatype]:
        return QUANTIZE_AGNOSTIC_OPERATIONS

    @staticmethod
    def pre_layer_target_type() -> TargetType:
        return OV_PRE_LAYER_TARGET_TYPE

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> OVTargetPoint:
        return OVTargetPoint(target_type, target_node_name, port_id)

    @staticmethod
    def is_node_with_weights(node: NNCFNode) -> bool:
        return node.layer_attributes and node.layer_attributes.constant_attributes

    @staticmethod
    def get_activations_port_id(node: NNCFNode, nncf_graph: NNCFGraph) -> int:
        weight_ports = node.layer_attributes.get_const_port_ids()
        activation_ports = [
            e.input_port_id for e in nncf_graph.get_input_edges(node) if e.input_port_id not in weight_ports
        ]

        if len(activation_ports) != 1:
            raise nncf.InternalError(f"Too many weight or activation ports for {node.node_name} node")
        return activation_ports[0]

    @staticmethod
    def get_abs_max_channel_collector(
        num_samples: int, stats_reduction_axes: Tuple[int], inplace: bool, branch_key: str
    ) -> TensorCollector:
        collector = TensorCollector()
        reducer = OVAbsMaxReducer(reduction_axes=stats_reduction_axes, inplace=inplace)
        aggregator = MaxAggregator(num_samples=num_samples)
        collector.register_statistic_branch(branch_key, reducer, aggregator)
        return collector

    @staticmethod
    def get_weight_value(node_with_weight: NNCFNode, model: ov.Model, nncf_graph: NNCFGraph) -> Tensor:
        port_id = OVSmoothQuantAlgoBackend.get_weight_tensor_port_id(node_with_weight)
        return Tensor(get_weight_value(node_with_weight, model, port_id))

    @staticmethod
    def get_weight_tensor_port_id(node: NNCFNode) -> int:
        const_ids = node.layer_attributes.get_const_port_ids()
        if len(const_ids) != 1:
            raise nncf.InternalError(f"Found more than 1 port for {node.node_name} node")
        return const_ids[0]

    @staticmethod
    def weight_update_command(node_with_weight: NNCFNode, weight_value: np.ndarray) -> OVWeightUpdateCommand:
        weight_port_id = OVSmoothQuantAlgoBackend.get_weight_tensor_port_id(node_with_weight)
        return OVCommandCreator.create_command_to_update_weight(node_with_weight, weight_value, weight_port_id)

    @staticmethod
    def scale_insertion_command(
        source_node: NNCFNode,
        scale_value: np.ndarray,
        source_output_port_id: int,
        nodes: List[NNCFNode],
        scale_node_name: str,
    ) -> OVMultiplyInsertionCommand:
        return OVCommandCreator.multiply_insertion_command(
            source_node, nodes, source_output_port_id, scale_value, scale_node_name
        )

    @staticmethod
    def get_activation_channel_axis(node: NNCFNode, port_id: int) -> int:
        channel_axis = 1

        if port_id > 1:
            raise nncf.InternalError(f"{node.metatype.name} can not take more than 2 input tensors.")

        if (
            node.metatype == OVMatMulMetatype
            and node.layer_attributes is not None
            and node.layer_attributes.input_attributes is not None
            and "transpose" in node.layer_attributes.input_attributes
        ):
            transpose = node.layer_attributes.input_attributes["transpose"]
            channel_axis = OVSmoothQuantAlgoBackend.calculate_port_based_channel_axis(port_id, transpose)

        return channel_axis

    @staticmethod
    def get_weight_channel_axis(node: NNCFNode) -> int:
        if node.metatype != OVMatMulMetatype:
            return 1

        weights_layout = get_linear_weights_layout_from_node(node)
        return weights_layout.index(OVLayoutElem.C_IN)

    @staticmethod
    def calculate_port_based_channel_axis(port_id: int, transpose: bool) -> int:
        return -2 + port_id if transpose else -1 - port_id

    @staticmethod
    def is_node_with_shared_weight(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        weight_port_id = OVSmoothQuantAlgoBackend.get_weight_tensor_port_id(node)
        weight_node = nncf_graph.get_input_edge_by_port_id(node, weight_port_id).from_node
        return len(nncf_graph.get_next_nodes(weight_node)) > 1

    @staticmethod
    def get_filter_fn_for_statistics(activation_port_id: int, algorithm_key: str) -> Callable[[StatisticPoint], bool]:
        def filter_func(point: StatisticPoint) -> bool:
            return (
                algorithm_key in point.algorithm_to_tensor_collectors
                and point.target_point.type == OV_PRE_LAYER_TARGET_TYPE
                and point.target_point.port_id == activation_port_id
            )

        return filter_func



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict

import tensorflow as tf

from nncf.api.compression import CompressionLoss


class TFZeroCompressionLoss(CompressionLoss):
    def calculate(self, *args, **kwargs) -> Any:
        return tf.constant(0.0)

    def load_state(self, state: Dict[str, Any]) -> None:
        pass

    def get_state(self) -> Dict[str, Any]:
        return {}



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import Dict, List

from torch.nn import Module

from nncf.common.logging import nncf_logger


class CallCountTracker:
    def __init__(self, name):
        self.name = name
        self.call_counts = {}

    def init_with_key_list(self, key_list: List):
        self.call_counts = {key: 0 for key in key_list}
        nncf_logger.debug(f"{self.name} tracker: registered {len(self.call_counts)} entries")

    def register_call(self, key, counts=None):
        if key not in self.call_counts:
            nncf_logger.debug(f"DEBUG: {self.name} tracker: called an unregistered module: {key}")
            return
        if counts is None:
            self.call_counts[key] += 1
        else:
            self.call_counts[key] = counts

    def get_never_called_keys(self) -> List[str]:
        return [k for k, v in self.call_counts.items() if v == 0]

    def get_overcalled_keys_with_call_counts(self) -> Dict[str, int]:
        return {k: v for k, v in self.call_counts.items() if v > 1}

    def get_total_call_count(self) -> int:
        if self.call_counts:
            return sum(self.call_counts.values())
        return 0

    def reset(self):
        for key in self.call_counts:
            self.call_counts[key] = 0


class DebugInterface:
    def pre_forward_actions(self, module: Module):
        raise NotImplementedError

    def post_forward_actions(self, module: Module):
        raise NotImplementedError

    def init_actual(self, owner_model):
        raise NotImplementedError


def debuggable_forward(forward_func):
    from nncf.torch.nncf_network import NNCFNetwork

    @functools.wraps(forward_func)
    def decorated(self: NNCFNetwork, *args, **kwargs):
        if hasattr(self, "nncf") and self.nncf.debug_interface is not None:
            self.nncf.debug_interface.pre_forward_actions(module=self)
        retval = forward_func(self, *args, **kwargs)
        if hasattr(self, "nncf") and self.nncf.debug_interface is not None:
            self.nncf.debug_interface.post_forward_actions(module=self)
        return retval

    return decorated


class CombinedDebugInterface(DebugInterface):
    def __init__(self):
        self._interfaces: List[DebugInterface] = []

    def add_interface(self, interface: "DebugInterface"):
        self._interfaces.append(interface)

    def init_actual(self, owner_model: "NNCFNetwork"):  # noqa: F821
        for interface in self._interfaces:
            interface.init_actual(owner_model)

    def pre_forward_actions(self, module: Module):
        for interface in self._interfaces:
            interface.pre_forward_actions(module)

    def post_forward_actions(self, module: Module):
        for interface in self._interfaces:
            interface.post_forward_actions(module)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from collections import Counter
from typing import Any, Dict, TypeVar, cast

from nncf.tensor import Tensor
from nncf.tensor import functions as fns

TensorType = TypeVar("TensorType")


class TensorStatistic(ABC):
    """Base class that stores statistic data"""

    TENSOR_STATISTIC_OUTPUT_KEY = "tensor_statistic_output"

    @staticmethod
    def tensor_eq(tensor1: Tensor, tensor2: Tensor, rtol: float = 1e-6) -> bool:
        return cast(bool, fns.allclose(tensor1, tensor2))

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        pass


class MinMaxTensorStatistic(TensorStatistic):
    MIN_STAT = "min_values"
    MAX_STAT = "max_values"

    def __init__(self, min_values: Tensor, max_values: Tensor):
        self.min_values = min_values
        self.max_values = max_values

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, MinMaxTensorStatistic):
            return False
        return self.tensor_eq(self.min_values, other.min_values) and self.tensor_eq(self.max_values, other.max_values)


class MeanTensorStatistic(TensorStatistic):
    MEAN_STAT = "mean_values"
    SHAPE_STAT = "shape"

    """
    Base class for the statistics that collects as mean per-axis
    """

    def __init__(self, mean_values: Tensor, shape: Tensor):
        """
        :param mean_values: Collected mean per-axis values.
        :param shape: The shape of the collected statistics.
        """
        self.mean_values = mean_values
        self.shape = shape

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, MeanTensorStatistic):
            return False
        return self.tensor_eq(self.mean_values, other.mean_values) and self.tensor_eq(self.shape, other.shape)


class MedianMADTensorStatistic(TensorStatistic):
    MEDIAN_VALUES_STAT = "median_values"
    MAD_VALUES_STAT = "mad_values"

    def __init__(self, median_values: Tensor, mad_values: Tensor):
        self.median_values = median_values
        self.mad_values = mad_values

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, MedianMADTensorStatistic):
            return False
        return self.tensor_eq(self.median_values, other.median_values) and self.tensor_eq(
            self.mad_values, other.mad_values
        )


class PercentileTensorStatistic(TensorStatistic):
    PERCENTILE_VS_VALUE_DICT = "percentile_vs_values_dict"

    def __init__(self, percentile_vs_values_dict: Dict[float, Any]):
        self.percentile_vs_values_dict = percentile_vs_values_dict

    def __eq__(self, other: Any, rtol: float = 1e-9) -> bool:
        if not isinstance(other, PercentileTensorStatistic):
            return False
        if Counter(self.percentile_vs_values_dict.keys()) != Counter(other.percentile_vs_values_dict.keys()):
            return False
        for pct in self.percentile_vs_values_dict:
            if not self.tensor_eq(self.percentile_vs_values_dict[pct], other.percentile_vs_values_dict[pct]):
                return False
        return True


class RawTensorStatistic(TensorStatistic):
    VALUES_STATS = "values"

    """
    Base class for the raw statistics, without any aggregation.
    """

    def __init__(self, values: Tensor):
        """
        :param values: Collected raw values.
        """
        self.values = values

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, RawTensorStatistic):
            return False
        return self.tensor_eq(self.values, other.values)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from nncf.tensorflow.graph.utils import get_original_name_and_instance_idx


def is_ignored(node_name, ignored_scopes):
    original_name, _ = get_original_name_and_instance_idx(node_name)
    return any(
        (
            re.fullmatch(ignored.replace("{re}", ""), original_name)
            if ignored.startswith("{re}")
            else ignored == original_name
        )
        for ignored in ignored_scopes
    )



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.fx

import nncf.torch.graph.operator_metatypes as om
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.experimental.torch.fx.groups import FX_OPERATORS_WITH_BIAS_METATYPES
from nncf.tensor import Tensor


# TODO(dlyakhov): Use torch.fx.graph.find_nodes method instead after
# torch version update (>= 2.4)
def get_graph_node_by_name(graph: torch.fx.Graph, name: str) -> torch.fx.Node:
    """
    Retrieves a node with the specified name from the grpah.
    Raises a runtime error if graph does not contain node with
    the given name.

    :param graph: Given torch fx graph.
    :param name: Target node name.
    :return: A graph node with the given name.
    """
    for node in graph.nodes:
        if node.name == name:
            return node
    raise RuntimeError(f"Node with name {name} is not found")


def get_tensor_constant_from_node(constant_node: torch.fx.Node, model: torch.fx.GraphModule) -> torch.nn.Parameter:
    """
    Retrieves tensor from the given constant node.

    :param constant_node: Given constant node.
    :param model: Given model.
    :return: Torch tensor referenced by the given constant node.
    """
    if constant_node is None:
        return None
    if constant_node.op != "get_attr":
        raise RuntimeError(f"Given node op == {constant_node.op}, but get_attr is expected.")
    target_atoms = constant_node.target.split(".")
    attr_itr = model
    for i, atom in enumerate(target_atoms):
        if not hasattr(attr_itr, atom):
            raise RuntimeError(f"Node referenced nonexistent target {'.'.join(target_atoms[:i])}")
        attr_itr = getattr(attr_itr, atom)
    return attr_itr


def is_node_with_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
    """
    Returns True if the node has a bias, False otherwise.

    :param node: Target node.
    :param nncf_graph: Target nncf_graph.
    :return: True if the node has a bias, False otherwise.
    """
    # Assumes that all biases were unfused
    if node.metatype not in FX_OPERATORS_WITH_BIAS_METATYPES or len(nncf_graph.get_input_edges(node)) != 3:
        return False
    const_node = nncf_graph.get_input_edge_by_port_id(node, 2).from_node
    return const_node.metatype is om.PTConstNoopMetatype


def get_bias_value(node: NNCFNode, nncf_graph: NNCFGraph, model: torch.fx.GraphModule) -> Tensor:
    """
    Retrieves the bias value from the given node.

    :param node: Target node.
    :param nncf_graph: Target nncf_graph.
    :param model: Target GraphModule.
    :return: Bias value of the given node.
    """
    bias_node = nncf_graph.get_input_edge_by_port_id(node, 2).from_node
    # TODO(dlyakhov): make a node_name_vs_node map to speed up the process
    graph_bias_const = get_graph_node_by_name(model.graph, bias_node.node_name)
    return Tensor(get_tensor_constant_from_node(graph_bias_const, model))



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Dict, List

import networkx as nx

from nncf.common.utils.dot_file_rw import write_dot_graph
from nncf.experimental.common.pruning.propagation_data import PropagationGroup


class BlockHierarchy:
    """
    Graph that represents the hierarchy of propagation blocks/groups
    """

    PROPAGATION_GROUP = "propagation_group"

    def __init__(self, root_groups: List[PropagationGroup]) -> None:
        """
        Creates hierarchy of propagation blocks/groups by traversing children in the given root groups.

        :param roots: list of the root groups
        :return: networkx graph that represents the hierarchy of propagation blocks/groups.
        """

        self._id_counter = 0
        self._graph = nx.DiGraph()
        self._visited_block_ids_map: Dict[int, int] = {}

        for root_group in root_groups:
            self._add_group_to_graph(root_group)
            self._id_counter += 1

    def get_groups_on_leaves(self) -> List[PropagationGroup]:
        """
        Returns the list of all propagation groups on the leaves.
        """
        groups = []
        for node_id, data in self._graph.nodes(data=True):
            is_leaf = self._graph.out_degree(node_id) == 0
            if is_leaf:
                groups.append(data[self.PROPAGATION_GROUP])
        return groups

    def visualize_graph(self, path: Path) -> None:
        out_graph = self._get_graph_for_visualization()
        write_dot_graph(out_graph, path)

    def _add_group_to_graph(self, parent_group: PropagationGroup) -> None:
        """
        Recursive helper to traverse children of the given PropagationBlock with adding them to the graph.

        :param parent_group: current group for traversing.
        """
        parent_graph_id = str(self._id_counter)
        is_leaf = not parent_group.has_children()
        attrs = {self.PROPAGATION_GROUP: parent_group}
        self._graph.add_node(parent_graph_id, **attrs)
        self._visited_block_ids_map[id(parent_group)] = parent_graph_id
        if not is_leaf:
            for child_group in parent_group.get_children():
                child_id = id(child_group)
                if child_id not in self._visited_block_ids_map:
                    self._id_counter += 1
                    child_graph_id = str(self._id_counter)
                    self._visited_block_ids_map[id(child_group)] = child_graph_id
                    self._graph.add_edge(parent_graph_id, child_graph_id)
                    self._add_group_to_graph(child_group)
                else:
                    child_graph_id = self._visited_block_ids_map[child_id]
                    self._graph.add_edge(parent_graph_id, child_graph_id)

    def _get_graph_for_visualization(self) -> nx.DiGraph:
        """
        :return: A user-friendly graph .dot file, making it easier to debug the block hierarchy.
        """
        out_graph = nx.DiGraph()
        for node_id, node_data in self._graph.nodes(data=True):
            group = node_data[self.PROPAGATION_GROUP]
            is_leaf = not group.has_children()
            color = "grey"
            if is_leaf:
                color = "red" if group.is_invalid else "green"
            out_graph.add_node(node_id, label=f'"{str(group)}"', color=color, style="filled")

        for u, v in self._graph.edges:
            out_graph.add_edge(u, v)

        return out_graph



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module implements selected functions from the `torch` module, excluding the `hand_function` mechanism.

It processes inner functions to handle exception hooks and graph analysis. The implementation is designed
to support custom handling of inner function exceptions for specific functions.
"""

import math
from typing import Any, Callable, Dict, Optional, Tuple, Union

import torch
from torch.nn.functional import _canonical_mask
from torch.nn.functional import _in_projection  # type:ignore[attr-defined]
from torch.nn.functional import _in_projection_packed  # type:ignore[attr-defined]
from torch.nn.functional import _mha_shape_check  # type:ignore[attr-defined]
from torch.nn.functional import _none_or_dtype
from torch.nn.functional import _verify_batch_size  # type:ignore[attr-defined]
from torch.nn.functional import dropout
from torch.nn.functional import linear
from torch.nn.functional import pad
from torch.nn.functional import scaled_dot_product_attention
from torch.nn.functional import softmax

Tensor = torch.Tensor


def multi_head_attention_forward(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    embed_dim_to_check: int,
    num_heads: int,
    in_proj_weight: Optional[Tensor],
    in_proj_bias: Optional[Tensor],
    bias_k: Optional[Tensor],
    bias_v: Optional[Tensor],
    add_zero_attn: bool,
    dropout_p: float,
    out_proj_weight: Tensor,
    out_proj_bias: Optional[Tensor],
    training: bool = True,
    key_padding_mask: Optional[Tensor] = None,
    need_weights: bool = True,
    attn_mask: Optional[Tensor] = None,
    use_separate_proj_weight: bool = False,
    q_proj_weight: Optional[Tensor] = None,
    k_proj_weight: Optional[Tensor] = None,
    v_proj_weight: Optional[Tensor] = None,
    static_k: Optional[Tensor] = None,
    static_v: Optional[Tensor] = None,
    average_attn_weights: bool = True,
    is_causal: bool = False,
) -> Tuple[Tensor, Optional[Tensor]]:

    is_batched = _mha_shape_check(query, key, value, key_padding_mask, attn_mask, num_heads)

    if not is_batched:
        query = query.unsqueeze(1)
        key = key.unsqueeze(1)
        value = value.unsqueeze(1)
        if key_padding_mask is not None:
            key_padding_mask = key_padding_mask.unsqueeze(0)

    tgt_len, bsz, embed_dim = query.shape
    src_len, _, _ = key.shape

    key_padding_mask = _canonical_mask(
        mask=key_padding_mask,
        mask_name="key_padding_mask",
        other_type=_none_or_dtype(attn_mask),
        other_name="attn_mask",
        target_type=query.dtype,
    )

    if is_causal and attn_mask is None:
        raise RuntimeError(
            "Need attn_mask if specifying the is_causal hint. "
            "You may use the Transformer module method "
            "`generate_square_subsequent_mask` to create this mask."
        )

    if is_causal and key_padding_mask is None and not need_weights:

        attn_mask = None
    else:
        attn_mask = _canonical_mask(
            mask=attn_mask,
            mask_name="attn_mask",
            other_type=None,
            other_name="",
            target_type=query.dtype,
            check_other=False,
        )

        if key_padding_mask is not None:

            is_causal = False

    assert (
        embed_dim == embed_dim_to_check
    ), f"was expecting embedding dimension of {embed_dim_to_check}, but got {embed_dim}"
    if isinstance(embed_dim, torch.Tensor):
        head_dim = embed_dim.div(num_heads, rounding_mode="trunc")
    else:
        head_dim = embed_dim // num_heads
    assert head_dim * num_heads == embed_dim, f"embed_dim {embed_dim} not divisible by num_heads {num_heads}"
    if use_separate_proj_weight:
        assert (
            key.shape[:2] == value.shape[:2]
        ), f"key's sequence and batch dims {key.shape[:2]} do not match value's {value.shape[:2]}"
    else:
        assert key.shape == value.shape, f"key shape {key.shape} does not match value shape {value.shape}"

    if not use_separate_proj_weight:
        assert in_proj_weight is not None, "use_separate_proj_weight is False but in_proj_weight is None"
        q, k, v = _in_projection_packed(query, key, value, in_proj_weight, in_proj_bias)
    else:
        assert q_proj_weight is not None, "use_separate_proj_weight is True but q_proj_weight is None"
        assert k_proj_weight is not None, "use_separate_proj_weight is True but k_proj_weight is None"
        assert v_proj_weight is not None, "use_separate_proj_weight is True but v_proj_weight is None"
        if in_proj_bias is None:
            b_q = b_k = b_v = None
        else:
            b_q, b_k, b_v = in_proj_bias.chunk(3)
        q, k, v = _in_projection(query, key, value, q_proj_weight, k_proj_weight, v_proj_weight, b_q, b_k, b_v)

    if attn_mask is not None:
        if attn_mask.dim() == 2:
            correct_2d_size = (tgt_len, src_len)
            if attn_mask.shape != correct_2d_size:
                raise RuntimeError(
                    f"The shape of the 2D attn_mask is {attn_mask.shape}, but should be {correct_2d_size}."
                )
            attn_mask = attn_mask.unsqueeze(0)
        elif attn_mask.dim() == 3:
            correct_3d_size = (bsz * num_heads, tgt_len, src_len)
            if attn_mask.shape != correct_3d_size:
                raise RuntimeError(
                    f"The shape of the 3D attn_mask is {attn_mask.shape}, but should be {correct_3d_size}."
                )
        else:
            raise RuntimeError(f"attn_mask's dimension {attn_mask.dim()} is not supported")

    if bias_k is not None and bias_v is not None:
        assert static_k is None, "bias cannot be added to static key."
        assert static_v is None, "bias cannot be added to static value."
        k = torch.cat([k, bias_k.repeat(1, bsz, 1)])
        v = torch.cat([v, bias_v.repeat(1, bsz, 1)])
        if attn_mask is not None:
            attn_mask = pad(attn_mask, (0, 1))
        if key_padding_mask is not None:
            key_padding_mask = pad(key_padding_mask, (0, 1))
    else:
        assert bias_k is None
        assert bias_v is None

    q = q.view(tgt_len, bsz * num_heads, head_dim).transpose(0, 1)
    if static_k is None:
        k = k.view(k.shape[0], bsz * num_heads, head_dim).transpose(0, 1)
    else:
        assert (
            static_k.size(0) == bsz * num_heads
        ), f"expecting static_k.size(0) of {bsz * num_heads}, but got {static_k.size(0)}"
        assert static_k.size(2) == head_dim, f"expecting static_k.size(2) of {head_dim}, but got {static_k.size(2)}"
        k = static_k
    if static_v is None:
        v = v.view(v.shape[0], bsz * num_heads, head_dim).transpose(0, 1)
    else:
        assert (
            static_v.size(0) == bsz * num_heads
        ), f"expecting static_v.size(0) of {bsz * num_heads}, but got {static_v.size(0)}"
        assert static_v.size(2) == head_dim, f"expecting static_v.size(2) of {head_dim}, but got {static_v.size(2)}"
        v = static_v

    if add_zero_attn:
        zero_attn_shape = (bsz * num_heads, 1, head_dim)
        k = torch.cat([k, torch.zeros(zero_attn_shape, dtype=k.dtype, device=k.device)], dim=1)
        v = torch.cat([v, torch.zeros(zero_attn_shape, dtype=v.dtype, device=v.device)], dim=1)
        if attn_mask is not None:
            attn_mask = pad(attn_mask, (0, 1))
        if key_padding_mask is not None:
            key_padding_mask = pad(key_padding_mask, (0, 1))

    src_len = k.size(1)

    if key_padding_mask is not None:
        assert key_padding_mask.shape == (
            bsz,
            src_len,
        ), f"expecting key_padding_mask shape of {(bsz, src_len)}, but got {key_padding_mask.shape}"
        key_padding_mask = (
            key_padding_mask.view(bsz, 1, 1, src_len).expand(-1, num_heads, -1, -1).reshape(bsz * num_heads, 1, src_len)
        )
        if attn_mask is None:
            attn_mask = key_padding_mask
        else:
            attn_mask = attn_mask + key_padding_mask

    if not training:
        dropout_p = 0.0

    if need_weights:
        B, Nt, E = q.shape  # noqa: F841
        q_scaled = q * math.sqrt(1.0 / float(E))

        assert not (is_causal and attn_mask is None), "FIXME: is_causal not implemented for need_weights"

        if attn_mask is not None:
            attn_output_weights = torch.baddbmm(attn_mask, q_scaled, k.transpose(-2, -1))
        else:
            attn_output_weights = torch.bmm(q_scaled, k.transpose(-2, -1))
        attn_output_weights = softmax(attn_output_weights, dim=-1)
        if dropout_p > 0.0:
            attn_output_weights = dropout(attn_output_weights, p=dropout_p)

        attn_output = torch.bmm(attn_output_weights, v)

        attn_output = attn_output.transpose(0, 1).contiguous().view(tgt_len * bsz, embed_dim)
        attn_output = linear(attn_output, out_proj_weight, out_proj_bias)
        attn_output = attn_output.view(tgt_len, bsz, attn_output.size(1))

        attn_output_weights = attn_output_weights.view(bsz, num_heads, tgt_len, src_len)
        if average_attn_weights:
            attn_output_weights = attn_output_weights.mean(dim=1)

        if not is_batched:
            attn_output = attn_output.squeeze(1)
            attn_output_weights = attn_output_weights.squeeze(0)
        return attn_output, attn_output_weights
    else:
        if attn_mask is not None:
            if attn_mask.size(0) == 1 and attn_mask.dim() == 3:
                attn_mask = attn_mask.unsqueeze(0)
            else:
                attn_mask = attn_mask.view(bsz, num_heads, -1, src_len)

        q = q.view(bsz, num_heads, tgt_len, head_dim)
        k = k.view(bsz, num_heads, src_len, head_dim)
        v = v.view(bsz, num_heads, src_len, head_dim)

        attn_output = scaled_dot_product_attention(q, k, v, attn_mask, dropout_p, is_causal)
        attn_output = attn_output.permute(2, 0, 1, 3).contiguous().view(bsz * tgt_len, embed_dim)

        attn_output = linear(attn_output, out_proj_weight, out_proj_bias)
        attn_output = attn_output.view(tgt_len, bsz, attn_output.size(1))
        if not is_batched:
            attn_output = attn_output.squeeze(1)
        return attn_output, None


def relu(input: Tensor, inplace: bool = False) -> Tensor:
    if inplace:
        result = torch.relu_(input)
    else:
        result = torch.relu(input)
    return result


def batch_norm(
    input: Tensor,
    running_mean: Optional[Tensor],
    running_var: Optional[Tensor],
    weight: Optional[Tensor] = None,
    bias: Optional[Tensor] = None,
    training: bool = False,
    momentum: float = 0.1,
    eps: float = 1e-5,
) -> Tensor:
    if training:
        _verify_batch_size(input.size())

    return torch.batch_norm(
        input, weight, bias, running_mean, running_var, training, momentum, eps, torch.backends.cudnn.enabled
    )


MAP_HANDLER_TO_INNER_FUNCTION: Dict[Callable[..., Any], Callable[..., Any]] = {
    torch.nn.functional.relu: relu,
    torch.nn.functional.multi_head_attention_forward: multi_head_attention_forward,
    torch.nn.functional.batch_norm: batch_norm,
}


def get_handle_inner_function(fn: Callable[..., Any]) -> Union[Callable[..., Any], None]:
    """
    Returns the corresponding function to process inner functions.

    :param fn: The function for which the handler is needed.
    :return: The function without handle_function, or None if no handler is found.
    """
    return MAP_HANDLER_TO_INNER_FUNCTION.get(fn)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import defaultdict
from typing import Dict, List, Tuple, Union

import torch
import torch.fx
from torch.ao.quantization.quantizer import Quantizer as TorchAOQuantizer
from torch.ao.quantization.quantizer.quantizer import QuantizationSpec
from torch.ao.quantization.quantizer.quantizer import QuantizationSpecBase
from torch.ao.quantization.quantizer.quantizer import SharedQuantizationSpec

import nncf
from nncf.common.graph.graph import NNCFGraph
from nncf.common.logging import nncf_logger
from nncf.common.quantization.quantizer_setup import ActivationQuantizationInsertionPoint
from nncf.common.quantization.quantizer_setup import QuantizationPointBase
from nncf.common.quantization.quantizer_setup import SingleConfigQuantizationPoint
from nncf.common.quantization.quantizer_setup import SingleConfigQuantizerSetup
from nncf.common.quantization.quantizer_setup import WeightQuantizationInsertionPoint
from nncf.common.quantization.structs import QuantizationScheme as QuantizationMode
from nncf.common.quantization.structs import QuantizerConfig
from nncf.experimental.quantization.quantizer import Quantizer
from nncf.experimental.torch.fx.nncf_graph_builder import GraphConverter

EdgeOrNode = Union[Tuple[torch.fx.Node, torch.fx.Node]]


class TorchAOQuantizerAdapter(Quantizer):
    """
    Implementation of the NNCF Quantizer interface for any given torch.ao quantizer.
    """

    def __init__(self, quantizer: TorchAOQuantizer):
        self._quantizer = quantizer

    def transform_prior_quantization(self, model: torch.fx.GraphModule) -> torch.fx.GraphModule:
        return self._quantizer.transform_for_annotation(model)

    def get_quantization_setup(self, model: torch.fx.GraphModule, nncf_graph: NNCFGraph) -> SingleConfigQuantizerSetup:
        # Save model and nodes meta before the annotation
        original_meta = model.meta.copy()
        node_name_vs_meta = {}
        with torch.no_grad():
            for node in model.graph.nodes:
                node_name_vs_meta[node.name] = node.meta.copy()

        annotated_model = self._quantizer.annotate(model)
        self._quantizer.validate(annotated_model)
        quantizer_setup = self.get_quantizer_config_from_annotated_model(annotated_model)

        # Recover original meta
        model.meta = original_meta
        for node in model.graph.nodes:
            node.meta = node_name_vs_meta[node.name]

        return quantizer_setup

    @staticmethod
    def _get_quantization_points(
        from_node: torch.fx.Node,
        to_nodes: List[torch.fx.Node],
        anotated_model: torch.fx.GraphModule,
        qconfig: QuantizerConfig,
    ) -> List[QuantizationPointBase]:
        to_n = to_nodes[0]
        if from_node.op == "get_attr":
            _, metatype = GraphConverter.get_node_type_and_metatype(to_n, anotated_model)
            # Check that the constant is placed on the actual weight port, as it is possible for
            # activations to be a constant as well.
            if TorchAOQuantizerAdapter._get_node_args(to_n).index(from_node) in metatype.weight_port_ids:
                qip = WeightQuantizationInsertionPoint(to_n.name)
                return [SingleConfigQuantizationPoint(qip, qconfig, [x.name for x in to_nodes])]

        if len(from_node.users) == len(to_nodes):
            qip = ActivationQuantizationInsertionPoint(from_node.name)
            return [SingleConfigQuantizationPoint(qip, qconfig, [x.name for x in to_nodes])]

        qps = []
        for to_n_ in to_nodes:
            input_port_id = to_n_.args.index(from_node)
            qip = ActivationQuantizationInsertionPoint(to_n_.name, input_port_id)
            qp = SingleConfigQuantizationPoint(qip, qconfig, [to_n_.name])
            qps.append(qp)
        return qps

    @staticmethod
    def _get_node_args(node: torch.fx.Node):
        if node.target == torch.ops.aten.cat.default:
            return node.args[0]
        return node.args

    @staticmethod
    def get_quantizer_config_from_annotated_model(anotated_model: torch.fx.GraphModule) -> SingleConfigQuantizerSetup:
        edge_or_node_to_qspec = _get_edge_or_node_to_qspec(anotated_model)

        q_map = defaultdict(list)
        for edge, qspec in edge_or_node_to_qspec.items():
            if not isinstance(edge, tuple):
                continue
            from_n, to_n = edge
            q_map[from_n].append(to_n)

        q_setup = SingleConfigQuantizerSetup()
        for from_n, to_nodes in q_map.items():
            to_n = to_nodes[0]
            qspec = edge_or_node_to_qspec[(from_n, to_n)]
            if qspec is None:
                continue
            if isinstance(qspec, QuantizationSpec):
                if qspec.qscheme in [torch.per_channel_affine, torch.per_channel_symmetric]:
                    per_channel = True
                elif qspec.qscheme in [torch.per_tensor_affine, torch.per_tensor_symmetric]:
                    per_channel = False
                else:
                    raise nncf.InternalError(f"Unknown qscheme: {qspec.qscheme}")
                signed = qspec.dtype is torch.int8
                mode = (
                    QuantizationMode.SYMMETRIC
                    if qspec.qscheme in [torch.per_channel_symmetric, torch.per_tensor_symmetric]
                    else QuantizationMode.ASYMMETRIC
                )
                qconfig = QuantizerConfig(mode=mode, signedness_to_force=signed, per_channel=per_channel)

                qps = TorchAOQuantizerAdapter._get_quantization_points(from_n, to_nodes, anotated_model, qconfig)
                for qp in qps:
                    q_setup.add_independent_quantization_point(qp)

            elif isinstance(qspec, SharedQuantizationSpec):
                # TODO(dlyakhov): Support SharedQuantizationSpec
                nncf_logger.warning(
                    "SharedQuantizationSpec is not supported yet;" f" edges {from_n} -> {to_nodes} won't be quantized."
                )
            else:
                raise nncf.InternalError(f"Unknown torch.ao quantization spec: {qspec}")

        return q_setup


def _get_edge_or_node_to_qspec(
    model: torch.fx.GraphModule,
) -> Dict[EdgeOrNode, QuantizationSpecBase]:
    """
    Get a map from EdgeOrNode to quantization spec based on annotations on the nodes.

    :param model: torch.fx.GraphModule instance.
    :return: A map from EdgeOrNode to quantization spec based on annotations on the nodes.
    """
    edge_or_node_to_qspec: Dict[EdgeOrNode, QuantizationSpecBase] = {}
    for n in model.graph.nodes:
        if hasattr(n, "meta") and "quantization_annotation" in n.meta:
            qa = n.meta["quantization_annotation"]
            for input_to_n, qspec in qa.input_qspec_map.items():
                input_edge = (input_to_n, n)
                edge_or_node_to_qspec[input_edge] = qspec
            if qa.output_qspec is not None:
                output_node = n
                qspec = qa.output_qspec
                edge_or_node_to_qspec[output_node] = qspec
    return edge_or_node_to_qspec



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import OrderedDict
from functools import partial
from itertools import islice
from typing import Dict

from nncf.torch.nested_objects_traversal import InputIndexEntry
from nncf.torch.nested_objects_traversal import NestedObjectIndex


def nth(iterable, n, default=None):
    return next(islice(iterable, n, None), default)


class OperatorInput:
    def __init__(self, op_args, op_kwargs):
        self.op_args = op_args
        self.op_kwargs = op_kwargs
        self._index: Dict[int, InputIndexEntry] = OrderedDict()

        op_args_index_entries = NestedObjectIndex(self.op_args, previous_level_setter=partial(setattr, self, "op_args"))
        op_kwargs_index_entries = NestedObjectIndex(self.op_kwargs)

        self._index = {
            idx: entry
            for idx, entry in enumerate(
                op_args_index_entries.get_flat_nested_obj_indexing()
                + op_kwargs_index_entries.get_flat_nested_obj_indexing()
            )
        }

    def __iter__(self):
        return iter(self._index.values())

    def __getitem__(self, n):
        return self._index[n].getter()

    def __setitem__(self, n, value):
        self._index[n].setter(value)

    def __len__(self):
        return len(self._index)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum

from nncf.common.utils.api_marker import api


class StrEnum(str, Enum):
    def __str__(self) -> str:
        return str(self.value)


@api(canonical_alias="nncf.TargetDevice")
class TargetDevice(StrEnum):
    """
    Target device architecture for compression.

    Compression will take into account the value of this parameter in order to obtain the best performance
    for this type of device.
    """

    ANY = "ANY"
    CPU = "CPU"
    GPU = "GPU"
    NPU = "NPU"
    CPU_SPR = "CPU_SPR"


@api(canonical_alias="nncf.ModelType")
class ModelType(StrEnum):
    """
    Describes the model type the specificity of which will be taken into account during compression.

    :param TRANSFORMER: Transformer-based models
        (https://arxiv.org/pdf/1706.03762.pdf)
    """

    TRANSFORMER = "transformer"


@api(canonical_alias="nncf.DropType")
class DropType(StrEnum):
    """
    Describes the accuracy drop type, which determines how the accuracy drop between
    the original model and the compressed model is calculated.

    :param ABSOLUTE: The accuracy drop is calculated as the absolute drop with respect
        to the results of the original model.
    :param RELATIVE: The accuracy drop is calculated relative to the results of
        the original model.
    """

    ABSOLUTE = "absolute"
    RELATIVE = "relative"


@api(canonical_alias="nncf.CompressWeightsMode")
class CompressWeightsMode(StrEnum):
    """
    Defines a mode for weight compression.
    :param INT8_SYM: Stands for 8-bit integer symmetric quantization of all weights.
        Weights are quantized symmetrically without zero point.
        https://github.com/openvinotoolkit/nncf/blob/develop/docs/usage/training_time_compression/other_algorithms/LegacyQuantization.md#symmetric-quantization
    :param INT8_ASYM: The same as INT8_SYM mode, but weights are quantized to a primary precision asymmetrically
        with a typical non-fixed zero point.
        https://github.com/openvinotoolkit/nncf/blob/develop/docs/compression_algorithms/Quantization.md#asymmetric-quantization
    :param INT4_SYM: Stands for a mixed-precision weights quantization with 4-bit integer as a primary precision.
        Weights are quantized to a primary precision symmetrically without zero point.
        All embeddings and the last layer are always compressed to a backup precision, which is INT8_ASYM,
        by default. All others are quantized whether to 4-bit integer or to a backup precision depending on
        criteria and the given ratio.
        https://github.com/openvinotoolkit/nncf/blob/develop/docs/usage/training_time_compression/other_algorithms/LegacyQuantization.md#symmetric-quantization
    :param INT4_ASYM: The same as INT4_SYM mode, but weights are quantized to a primary precision asymmetrically
        with a typical non-fixed zero point.
        https://github.com/openvinotoolkit/nncf/blob/develop/docs/usage/training_time_compression/other_algorithms/LegacyQuantization.md#asymmetric-quantization
    :param NF4: The the same as INT4_SYM mode, but primary precision is NF4 data type without zero point.
    :param INT8: Mode is deprecated and will be removed in future releases. Please use `INT8_ASYM` instead.
    :param E2M1: FP4 format from "OCP Microscaling Formats (MX) Specification" Version 1.0.
    """

    INT8_SYM = "int8_sym"
    INT8_ASYM = "int8_asym"
    INT4_SYM = "int4_sym"
    INT4_ASYM = "int4_asym"
    NF4 = "nf4"
    INT8 = "int8"  # Deprecated mode
    E2M1 = "e2m1"


@api(canonical_alias="nncf.BackupMode")
class BackupMode(StrEnum):
    """
    Defines a backup mode for weight compression.
    :param NONE: Stands for original floating-point precision of the model weights.
        In this mode, weights are retained in their original precision without any quantization.
    :param INT8_SYM: Stands for 8-bit integer symmetric quantization without zero point.
        https://github.com/openvinotoolkit/nncf/blob/develop/docs/usage/training_time_compression/other_algorithms/LegacyQuantization.md#symmetric-quantization
    :param INT8_ASYM: Stands for 8-bit integer asymmetric quantization with a typical non-fixed zero point.
        https://github.com/openvinotoolkit/nncf/blob/develop/docs/compression_algorithms/Quantization.md#asymmetric-quantization
    """

    NONE = "none"
    INT8_SYM = "int8_sym"
    INT8_ASYM = "int8_asym"


@api(canonical_alias="nncf.SensitivityMetric")
class SensitivityMetric(StrEnum):
    """
    Defines a sensitivity metric for assigning quantization precision to layers. In order to
        preserve the accuracy of the model, the more sensitive layers receives a higher precision.

    :param WEIGHT_QUANTIZATION_ERROR: The inverted 8-bit quantization noise. Weights with highest value
        of this metric can be accurately quantized channel-wise to 8-bit. The idea is to leave these weights in 8bit,
        and quantize the rest of layers to 4-bit group-wise. Since group-wise is more accurate than per-channel,
        accuracy should not degrade.
    :param HESSIAN_INPUT_ACTIVATION: The average Hessian trace of weights with respect to the layer-wise quantization
        error multiplied by L2 norm of 8-bit quantization noise.
    :param MEAN_ACTIVATION_VARIANCE: The mean variance of the layers' inputs
        multiplied by inverted 8-bit quantization noise.
    :param MAX_ACTIVATION_VARIANCE: The maximum variance of the layers' inputs
        multiplied by inverted 8-bit quantization noise.
    :param MEAN_ACTIVATION_MAGNITUDE: The mean magnitude of the layers' inputs
        multiplied by inverted 8-bit quantization noise.
    """

    WEIGHT_QUANTIZATION_ERROR = "weight_quantization_error"
    HESSIAN_INPUT_ACTIVATION = "hessian_input_activation"
    MEAN_ACTIVATION_VARIANCE = "mean_activation_variance"
    MAX_ACTIVATION_VARIANCE = "max_activation_variance"
    MEAN_ACTIVATION_MAGNITUDE = "mean_activation_magnitude"


@api(canonical_alias="nncf.QuantizationMode")
class QuantizationMode(StrEnum):
    """
    Defines special modes.
    Currently contains only FP8-related modes (https://arxiv.org/pdf/2209.05433.pdf).

    :param FP8_E4M3: Mode with 4-bit exponent and 3-bit mantissa.
    :param FP8_E5M2: Mode with 5-bit exponent and 2-bit mantissa.
    """

    FP8_E4M3 = "fp8_e4m3"
    FP8_E5M2 = "fp8_e5m2"



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import pandas as pd

import nncf
from nncf.common.logging import nncf_logger
from nncf.common.utils.debug import DEBUG_LOG_DIR
from nncf.common.utils.debug import is_debug
from nncf.experimental.common.tensor_statistics.statistics import WCTensorStatistic
from nncf.parameters import CompressWeightsMode
from nncf.quantization.advanced_parameters import AdvancedLoraCorrectionParameters
from nncf.quantization.algorithms.weight_compression.activation_stats import process_stats
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionConfig
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionParameters
from nncf.quantization.algorithms.weight_compression.weight_lowering import CompressedWeight
from nncf.quantization.algorithms.weight_compression.weight_lowering import do_int_dequantization
from nncf.quantization.algorithms.weight_compression.weight_lowering import do_nf4_dequantization
from nncf.quantization.algorithms.weight_compression.weight_lowering import do_nf4_quantization
from nncf.tensor import Tensor
from nncf.tensor import functions as fns
from nncf.tensor.definitions import TensorDataType


class DebugInterface:
    """
    Utility class to collect and dump debug information of the Lora Correction algorithm.
    """

    def __init__(self):
        self._noise_per_layer = {}

    def add_noises(self, layer_name: str, value: float):
        self._noise_per_layer[layer_name] = value

    def dump_data(self):
        if not self._noise_per_layer:
            return
        dump_dir = Path(DEBUG_LOG_DIR) / "lora"
        dump_dir.mkdir(parents=True, exist_ok=True)

        layer_dir = dump_dir / "per_layer"
        layer_dir.mkdir(parents=True, exist_ok=True)

        df = pd.DataFrame(self._noise_per_layer)
        losses_path = dump_dir / "noises.csv"
        nncf_logger.debug(f"Quantization noise through the correction process is saved to: {losses_path}")
        df.to_csv(losses_path)

        for name in df.columns:
            plt.plot(df[name])
            filename = name.replace("/", "_") + ".jpg"
            plt.savefig(layer_dir / filename)
            plt.clf()

        delta = df.iloc[0] - df.iloc[-1]
        nncf_logger.debug(f"Is quantization noise reduced for all layers: {all(delta > 0)}")

        _, ax = plt.subplots(1)
        ax.plot(delta)
        ax.set_xticklabels([])
        delta_path = dump_dir / "qnoise_change.jpg"
        nncf_logger.debug(f"Saving change in quantization noise for each layer to: {delta_path}")
        plt.savefig(delta_path)
        plt.clf()


class LoraCorrectionAlgorithm:
    """
    Contains implementation of LoRA Correction algorithm.

    The method reduces quantization noise after weight compression using low rank adapters.
    """

    def __init__(
        self, statistics: Dict[str, WCTensorStatistic], lora_correction_params: AdvancedLoraCorrectionParameters
    ):
        """
        :param statistics: Input activation statistics for each node.
        :param lora_correction_params: parameters to configure the algorithm.
        """
        self._statistics = statistics
        self._lora_correction_params = lora_correction_params
        self._debug_interface = DebugInterface() if is_debug() else None

    def __del__(self):
        if self._debug_interface is not None:
            self._debug_interface.dump_data()

    @property
    def use_int8_adapters(self) -> bool:
        return self._lora_correction_params.use_int8_adapters

    def is_applicable(self, wc_params: WeightCompressionParameters):
        return wc_params.compression_config.num_bits == 4

    def calculate_adapters(
        self, weight: Tensor, compressed_weight: CompressedWeight, wc_params: WeightCompressionParameters
    ) -> Tuple[Tensor, Tensor, List[float]]:
        """
        Calculates low rank matrices for a given original and compressed weights.

        :param weight: original floating-point weight matrix.
        :param compressed_weight: compressed weight matrix.
        :param wc_params: parameters of weight compression.
        :return: two low rank matrices in the order of execution of corresponding linear layers.
        """
        layer_name = wc_params.node_with_weight.node_name
        layer_statistics = self._statistics[layer_name]
        is_debug = self._debug_interface is not None
        lora_A, lora_B, mean_noises = self.calculate_low_rank_matrices(
            weight,
            compressed_weight,
            wc_params.compression_config,
            wc_params.reduction_axes,
            self._lora_correction_params,
            layer_statistics,
            is_debug,
        )
        if is_debug:
            self._debug_interface.add_noises(layer_name, mean_noises)
        return lora_A, lora_B

    @staticmethod
    def calculate_low_rank_matrices(
        weight: Tensor,
        compressed_weight: CompressedWeight,
        compression_config: WeightCompressionConfig,
        reduction_axes: Tuple[int, ...],
        lora_correction_params: AdvancedLoraCorrectionParameters,
        layer_statistics: WCTensorStatistic,
        is_debug: Optional[bool] = False,
    ):
        """
        Calculates low rank matrices for a given original and compressed weights.
        The low rank matrices obtained by applying singular value decomposition (SVD) with lower rank for the
        difference between original weight and fake-quantized ones.
        Then, an iterative algorithm refines them. It solves a system of linear equations alternately fixing
        one matrix, then another.

        :param weight: original floating-point weight matrix.
        :param compressed_weight: compressed weight matrix.
        :param compression_config: configuration of weight compression for the weight node.
        :param reduction_axes: axes along which different statistics reduced.
        :param lora_correction_params: parameters to configure the algorithm.
        :param layer_statistics: an object containing statistics for the layer.
        :param is_debug: whether to collect debug information, defaults to False.
        :return: two low rank matrices in the order of execution of corresponding linear layers and list of mean noises.
            Noises are collected from each step of the algorithm if debug was enabled.
            First low rank matrix has shape=[R, H], the second - [O, R], where R - rank, H/O - hidden/output dimension.
        """
        rank, num_iterations, apply_regularization, subset_size = (
            lora_correction_params.adapter_rank,
            lora_correction_params.num_iterations,
            lora_correction_params.apply_regularization,
            lora_correction_params.subset_size,
        )
        mode = compression_config.mode
        assert len(reduction_axes) == 1, "Assumed a single reduction axis"
        reduction_axis = reduction_axes[0] if compression_config.group_size != -1 else -1
        if mode in (CompressWeightsMode.INT4_SYM, CompressWeightsMode.INT4_ASYM):
            fq_weights = do_int_dequantization(
                compressed_weight.tensor,
                compressed_weight.scale,
                compressed_weight.zero_point,
                reduction_axis,
            )
        elif mode == CompressWeightsMode.NF4:
            indexes = do_nf4_quantization(compressed_weight.tensor, compressed_weight.scale, is_normalized_weight=True)
            fq_weights = do_nf4_dequantization(indexes, compressed_weight.scale, reduction_axis)
        else:
            raise nncf.InternalError(
                f"{mode.value} mode is invalid for Lora Correction algorithm. Supported modes: INT4_SYM, INT4_ASYM, NF4"
            )
        # fq_w + residual = w   =>  residual = w - fq_w
        svd_residual = fns.astype(weight - fq_weights, TensorDataType.float32)

        # O stands for output dimension, H - input dimension or hidden size, SS - samples size, R - rank.
        # reduction axes is all axes except output dimension in linear/conv layers.
        if reduction_axes[0] == 1:
            svd_residual = fns.transpose(svd_residual)
        residual = svd_residual.clone()  # [H, O]

        s, X = process_stats(layer_statistics, subset_size)  # [H], [H, SS]
        X = fns.transpose(X)  # [SS, H]
        if compression_config.group_size > 0:
            # Multiply residual of weights by maximum channel magnitude of activations normalized per quantization
            # group. As a consequence, weights corresponding to a "noisy" activations has a higher error to correct.
            # Empirically, it leads to a better accuracy.
            gs = compression_config.group_size
            n_gs = s.shape[0] // gs
            for i in range(n_gs):
                offset = i * gs
                denum = fns.sum(s[offset : offset + gs])
                s[offset : offset + gs] = s[offset : offset + gs] / denum
                denum = fns.max(s[offset : offset + gs])
                s[offset : offset + gs] = s[offset : offset + gs] / denum
            s = fns.expand_dims(s, 1)  # [H, 1]
            svd_residual = svd_residual * s  # [H, O]

        # Low-rank approximation.
        U_full, S_full, V_full = fns.linalg.svd(svd_residual, full_matrices=False)
        U = U_full[:, :rank]  # [H, R]
        S = fns.diag(S_full[:rank])  # [R, R]
        V = V_full[:rank, :]  # [R, O]
        V = S @ V  # [R, O]

        # An iterative algorithm for correction (refinement) of the low-rank adapters.
        mean_noises = []
        noise = X @ residual  # [SS, H] * [H, O] = [SS, O]
        for i in range(num_iterations):
            # Part 1: U is fixed, find V.
            XU = X @ U  # [SS, R]
            if not apply_regularization:
                # X @ U @ V = noise      ---> a @ x = b
                new_V = fns.linalg.lstsq(XU, noise, driver="gelsy")
            else:
                # 1) U @ V = res         <--- regularization
                # 2) X @ U @ V = noise
                # |U X @ U| @ V = |res noise|
                UXU = fns.concatenate([U, XU], axis=0)  # [H + SS, R]
                noiseR = fns.concatenate([residual, noise], axis=0)  # [H + SS, O]
                new_V = fns.linalg.lstsq(UXU, noiseR, driver="gelsy")
            if is_debug:
                if i == 0:
                    init_noise = noise
                    mean_noise_before_svd = fns.mean(fns.abs(init_noise)).item()
                    mean_noise_after_svd = fns.mean(fns.abs(init_noise - XU @ V)).item()
                    mean_noises.extend([mean_noise_before_svd, mean_noise_after_svd])
                mean_noise_after_correct = fns.mean(fns.abs(init_noise - XU @ new_V)).item()
                mean_noises.append(mean_noise_after_correct)
                nncf_logger.debug(
                    f"{i} Correction U: {mean_noise_before_svd}, {mean_noise_after_svd}, {mean_noise_after_correct}"
                )
            V = new_V

            # Part 2: V is fixed, find U.
            VI = fns.linalg.pinv(V)  # [O, R]
            noiseVI = noise @ VI  # [R, SS]
            if not apply_regularization:
                # VI = V^-1
                # 1) X @ U @ V = noise
                # 1) X @ U = noise @ VI     ---> a @ x = b
                U = fns.linalg.lstsq(X, noiseVI, driver="gelsy")
            else:
                # VI = V^-1, E - identity matrix
                # 1) U @ V = res           <--- regularization
                # 1) E @ U = res @ VI
                # 2) X @ U @ V = noise
                # 2) X @ U = noise @ VI
                # |E X| = | (UI @ res) (UI @ noise) |
                E = fns.eye(U.shape[0], backend=U.backend, dtype=U.dtype)  # [H, H]
                EX = fns.concatenate([E, X], axis=0)  # [H + SS, H]
                noiseR = fns.concatenate([residual @ VI, noiseVI], axis=0)  # [H + SS, R]
                U = fns.linalg.lstsq(EX, noiseR, driver="gelsy")
            if is_debug:
                mean_noise_after_correct = fns.mean(fns.abs(init_noise - X @ U @ V)).item()
                mean_noises.append(mean_noise_after_correct)
                nncf_logger.debug(
                    f"{i} Correction V: {mean_noise_before_svd}, {mean_noise_after_svd}, {mean_noise_after_correct}"
                )
        return fns.transpose(U), fns.transpose(V), mean_noises



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from contextlib import contextmanager
from typing import Callable, Iterator

from nncf.common.logging import nncf_logger


@contextmanager
def timer() -> Iterator[Callable[[], float]]:
    """
    Context manager to measure execution time.
    """
    start_time = end_time = time.perf_counter()
    yield lambda: end_time - start_time
    end_time = time.perf_counter()
    elapsed_time = end_time - start_time
    time_string = time.strftime("%H:%M:%S", time.gmtime(elapsed_time))
    nncf_logger.info(f"Elapsed Time: {time_string}")



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.graph.transformations.commands import TransformationType
from nncf.common.graph.transformations.layout import TransformationLayout


class TFTransformationLayoutV2(TransformationLayout):
    def register(self, transformation: TransformationCommand) -> None:
        """
        Registers the transformation command in the transformation layout.

        The `TFTransformationLayoutV2` is a simplified version of the
        `TransformationLayout` class where some redundant functionality
        was removed.

        :param transformation: The transformation command to be registered in
            the transformation layout.
        """
        if transformation.type == TransformationType.REMOVE:
            # TODO(andrey-churkin): Add support.
            pass
        elif transformation.type == TransformationType.INSERT:
            self._register_insertion_transformation(transformation)
        else:
            raise ValueError(f"Unknown type of transformation command: {transformation.type}")

    def _register_insertion_transformation(self, transformation: TransformationCommand) -> None:
        idx = None
        for curr_idx, t in enumerate(self.transformations):
            if t.check_command_compatibility(transformation):
                assert idx is None
                idx = curr_idx

        if idx is not None:
            self.transformations[idx] = self.transformations[idx].union(transformation)
        else:
            self.transformations.append(transformation)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from dataclasses import dataclass
from typing import Tuple

import nncf
from nncf.common.quantization.quantizers import calculate_asymmetric_level_ranges
from nncf.common.quantization.quantizers import calculate_symmetric_level_ranges
from nncf.common.quantization.quantizers import get_num_levels
from nncf.common.quantization.structs import QuantizationScheme as QuantizationMode
from nncf.common.quantization.structs import QuantizerConfig
from nncf.common.quantization.structs import QuantizerGroup
from nncf.experimental.common.tensor_statistics.statistics import MinMaxTensorStatistic
from nncf.quantization.advanced_parameters import FP8Type
from nncf.tensor import Tensor
from nncf.tensor import TensorDataType
from nncf.tensor import functions as fns


@dataclass
class FakeQuantizeParameters:
    """
    Class handles FakeQuantize layer attributes.

    :param input_low: Tensor with minimum limit for input value.
    :param input_high: Tensor with maximum limit for input value.
    :param output_low: Tensor with minimum quantized value.
    :param output_high: Tensor with maximum quantized value.
    :param levels: Number of quantization levels.
    """

    input_low: Tensor
    input_high: Tensor
    output_low: Tensor
    output_high: Tensor
    levels: int


@dataclass
class FakeConvertParameters:
    """
    Class handles FakeConvert layer attributes.

    :param scale: Tensor with the scale for input value.
    :param shift: Tensor with the shift for input value.
    :param destination_type: Destination type.
    """

    scale: Tensor
    shift: Tensor
    destination_type: FP8Type


def fix_zero_filters_symmetric(max_values: Tensor, eps: float = 0.01) -> Tensor:
    """
    Fixes zero filters for symmetric quantizer.

    :param max_values: Collected max values for the quantized insertion.
    :param eps: Correction coefficient.
    :return: Fixed the high quant number.
    """
    max_range = fns.max(max_values)
    lower_threshold = fns.maximum(max_range * eps, 8e-5)
    return fns.maximum(lower_threshold, max_values)


def fix_zero_filters_asymmetric(min_values: Tensor, max_values: Tensor, eps: float = 1e-8) -> Tuple[Tensor, Tensor]:
    """
    Fixes zero filters for asymmetric quantizer.

    :param min_values: Collected min values for the quantized insertion.
    :param max_values: Collected max values for the quantized insertion.
    :param eps: Correction coefficient.
    :return: A Tuple
        level_low - fixed the low quant number
        level_high - fixed the high quant number
    """
    ranges = max_values - min_values
    min_correction = 8e-4
    corrections = fns.where(ranges > min_correction, (fns.maximum(eps * ranges, ranges) - ranges) * 0.5, min_correction)

    level_low = min_values - corrections
    level_high = max_values + corrections
    return level_low, level_high


def tune_range(
    left_border: Tensor, right_border: Tensor, num_bits: int, unify_zp: bool = False
) -> Tuple[Tensor, Tensor]:
    """
    Tunes asymmetric quantization range to unify the zero point of all channels if `unify_zp` is True,
    or sets zero quant precisely to zero value otherwise.
    Function moves left or right borders to do this and doesn't make left border higher or
    right border lesser than its original values.

    :param left_border: Range left border.
    :param right_border: Range right border.
    :param num_bits: Number of bits to perform quantization.
    :param unify_zp: Whether to unify the zero point of all channels.
        If `True` calculates the average zero point of all channels and tune the max/min range.
    :return: Tuple with recomputed ranges.
    """
    level_high = 2**num_bits - 1

    if unify_zp:
        scale = (right_border - left_border) / level_high
        zero_point = -left_border / scale
        avg_zpts = fns.round(fns.mean(zero_point))
        qval = fns.ones_like(left_border) * avg_zpts
    else:
        s = level_high / (right_border - left_border)
        fval = -left_border * s
        qval = fns.round(fval)

    with warnings.catch_warnings():
        # If `qval` is 0 `rb` will equal `right_border`, and we don't want to show an unnecessary division by 0 warning
        # The same for (qval - level_high)
        warnings.simplefilter("ignore")
        ra_then_result = qval / (qval - level_high) * right_border
        rb_then_result = (qval - level_high) / qval * left_border
    ra = fns.where(qval < level_high, ra_then_result, left_border)
    rb = fns.where(qval > 0.0, rb_then_result, right_border)

    range_a = right_border - ra
    range_b = rb - left_border

    mask = fns.where(range_a > range_b, 1.0, 0.0)
    inv_mask = fns.abs(1.0 - mask)

    ra = mask * ra + inv_mask * left_border
    rb = inv_mask * rb + mask * right_border

    return ra, rb


def symmetric_range(
    min_values: Tensor,
    max_values: Tensor,
    levels: int,
    quantizer_config: QuantizerConfig,
    q_group: QuantizerGroup,
) -> Tuple[Tensor, Tensor]:
    """
    Calculates the numbers of the low and high quant for the symmetric quantization scheme.

    :param min_values: Collected min values for the quantized insertion.
    :param max_values: Collected max values for the quantized insertion.
    :param levels: Number of quantization levels.
    :param quantizer_config: Config of the quantization configuration.
    :return: A Tuple
        level_low - the low quant number
        level_high - the high quant number
    """
    level_high = fix_zero_filters_symmetric(max_values)
    if q_group == QuantizerGroup.WEIGHTS:
        level_low = -level_high
    else:
        signed = quantizer_config.signedness_to_force is True
        level_low = (
            fns.zeros_like(level_high)
            if fns.all(min_values >= 0) and not signed
            else -level_high * levels / (levels - 2)
        )

    level_low = level_low.astype(TensorDataType.float32)
    level_high = level_high.astype(TensorDataType.float32)
    return level_low, level_high


def asymmetric_range(
    min_values: Tensor,
    max_values: Tensor,
    quantizer_config: QuantizerConfig,
    q_group: QuantizerGroup,
    unify_zp: bool = False,
) -> Tuple[Tensor, Tensor]:
    """
    Calculates the numbers of the low and high quant for the asymmetric quantization scheme.

    :param min_values: Collected min values for the quantized insertion.
    :param max_values: Collected max values for the quantized insertion.
    :param quantizer_config: Config of the quantization configuration.
    :param unify_zp: Whether to unify the zero point.
        It is `True` for the per-tensor zero point constrain on KMB.
    :return: A Tuple
        level_low - the low quant number
        level_high - the high quant number
    """
    level_low, level_high = fix_zero_filters_asymmetric(min_values, max_values)
    level_low = fns.where(level_low < 0.0, level_low, 0.0)
    level_high = fns.where(level_high > 0.0, level_high, 0.0)

    if unify_zp and q_group == QuantizerGroup.ACTIVATIONS:
        raise NotImplementedError("Unified zero point is not supported for activations.")

    level_low, level_high = tune_range(level_low, level_high, quantizer_config.num_bits, unify_zp=unify_zp)
    level_low = level_low.astype(TensorDataType.float32)
    level_high = level_high.astype(TensorDataType.float32)
    return level_low, level_high


def get_quantizer_narrow_range(quantizer_config: QuantizerConfig, quant_group: QuantizerGroup) -> bool:
    """
    Returns narrow_range parameter: True if the range of quantized values is reduced by 1 compared to the
        naive case, False otherwise.

    :param quantizer_config: Config of the quantization configuration.
    :param quant_group: Group of the quantizer.
    :return: narrow_range parameter.
    """
    if quantizer_config.mode == QuantizationMode.SYMMETRIC:
        return quant_group == QuantizerGroup.WEIGHTS
    return False


def calculate_quantizer_parameters(
    statistics: MinMaxTensorStatistic,
    quantizer_config: QuantizerConfig,
    quant_group: QuantizerGroup,
    narrow_range: bool,
    half_range: bool = False,
) -> FakeQuantizeParameters:
    """
    Calculates FakeQuantize layer attributes for weight/activation quantizer.

    :param statistics: Collected statistics for the quantized insertion.
    :param quantizer_config: Config of the quantization configuration.
    :param quantizer_group: Group of the quantizer.
    :param narrow_range: True if the range of quantized values is reduced by 1 compared to the
        naive case, False otherwise.
    :param half_range: If True effectively only a half of a quantizer range is used.
        False - the full range is used.
    :return: Parameters of the FakeQuantize layer.
    """
    min_values = statistics.min_values.astype(TensorDataType.float32)
    max_values = statistics.max_values.astype(TensorDataType.float32)

    if half_range:
        input_low, input_high, levels = _calculate_scaled_parameters(
            min_values, max_values, quantizer_config, quant_group, narrow_range
        )
    else:
        num_bits = quantizer_config.num_bits
        if quantizer_config.mode == QuantizationMode.SYMMETRIC:
            level_low, level_high = calculate_symmetric_level_ranges(num_bits, signed=True, narrow_range=narrow_range)
            levels = get_num_levels(level_low, level_high)
            input_low, input_high = symmetric_range(min_values, max_values, levels, quantizer_config, quant_group)
        else:
            level_low, level_high = calculate_asymmetric_level_ranges(num_bits, narrow_range=narrow_range)
            levels = get_num_levels(level_low, level_high)
            input_low, input_high = asymmetric_range(min_values, max_values, quantizer_config, quant_group)

    if not quantizer_config.per_channel:
        input_low = fns.squeeze(input_low)
        input_high = fns.squeeze(input_high)

    output_low, output_high = input_low, input_high
    return FakeQuantizeParameters(input_low, input_high, output_low, output_high, levels)


def calculate_convert_parameters(
    statistics: MinMaxTensorStatistic,
    is_per_channel: False,
    destination_type: FP8Type = FP8Type.E4M3,
    activation_scale: float = 0.5,
) -> FakeConvertParameters:
    """
    Calculates FakeConvert layer attributes for weight/activation quantizer.

    :param statistics: Collected statistics for the quantized insertion.
    :param is_activation: Whether is for activation or weights.
    :param destination_type: Destination type that regulates maximum value for the formula.
    :param activation_scale: Factor for calculated activation scale.
    :return: Parameters of the FakeConvert layer.
    """

    destination_type_maximum = {FP8Type.E4M3: 448, FP8Type.E5M2: 57344}

    max_values = statistics.max_values
    min_values = statistics.min_values

    max_destination_value = destination_type_maximum[destination_type]
    tensor_dtype = fns.finfo(max_values)
    scale = max_destination_value / fns.maximum(max_values, fns.abs(min_values) + tensor_dtype.eps)
    if not is_per_channel:
        scale = fns.squeeze(activation_scale * scale)
    shift = fns.zeros_like(scale).astype(TensorDataType.float32)
    scale = scale.astype(TensorDataType.float32)
    return FakeConvertParameters(scale, shift, destination_type)


def _calculate_scaled_parameters(
    min_values: Tensor,
    max_values: Tensor,
    quantizer_config: QuantizerConfig,
    quant_group: QuantizerGroup,
    narrow_range: bool,
) -> Tuple[Tensor, Tensor, int]:
    """
    Calculates FakeQuantize layer attributes scaled to effectively use a half range of the quantization range.

    :param min_values: Minimum values of statistics for the quantizer.
    :param max_values: Maximum values of statistics for the quantizer.
    :param quantizer_config: Config of the quantization configuration.
    :param quantizer_group: Group of the quantizer.
    :param narrow_range: True if the range of quantized values is reduced by 1 compared to the
        naive case, False otherwise.
    :return: A Tuple
        input_low: Tensor with minimum limit for input value.
        input_high: Tensor with maximum limit for input value.
        levels: Number of quantization levels.
    """
    if quantizer_config.mode == QuantizationMode.ASYMMETRIC:
        raise nncf.ValidationError("half_range is only applied to symmetric quantization mode.")
    if quant_group != QuantizerGroup.WEIGHTS:
        raise nncf.ValidationError("half_range is only applied to weight quantizers.")

    num_bits = quantizer_config.num_bits
    level_low, level_high = calculate_symmetric_level_ranges(num_bits - 1, signed=True, narrow_range=False)
    levels = get_num_levels(level_low, level_high)
    input_low, input_high = symmetric_range(min_values, max_values, levels, quantizer_config, quant_group)

    export_level_low, export_level_high = calculate_symmetric_level_ranges(
        num_bits, signed=True, narrow_range=narrow_range
    )
    export_levels = get_num_levels(export_level_low, export_level_high)
    input_high *= (export_levels - 1) / (levels - 1)
    input_low *= (export_levels - 1) / (levels - 1)

    return input_low, input_high, export_levels


def calculate_scale_zero_point(
    input_low: Tensor,
    input_high: Tensor,
    level_low: int,
    level_high: int,
    narrow_range: bool,
) -> Tuple[Tensor, Tensor]:
    """
    Calculates scale and zero_point values for the quantizer.

    :param input_low: The minimum limit for an input value based on collected statistics.
    :param input_high: The maximum limit for an input value based on collected statistics.
    :param level_low: The minimum level in the integer range to quantize.
        The default is "0" for an unsigned range, and "-2^(bit-1)" for a signed one .
    :param level_high: The maximum level in the integer range to quantize.
        The default is "2^bits-1" for an unsigned range, and "2^(bit-1)-1" for a signed one.
    :param narrow_range: True if the range of quantized values is narrowed as compared to the
        naive case, False otherwise.
    :return: Scale and Zero point values.
    """
    levels = level_high - level_low if narrow_range else level_high - level_low + 1
    scale = ((input_high - input_low) / (levels - 1)).astype(TensorDataType.float32)
    eps = fns.finfo(scale).eps
    # NOTE: adding machine epsilon to avoid division by zero
    scale = fns.where(fns.abs(scale) < eps, eps, scale)
    expected_level_low = level_low + 1 if narrow_range else level_low
    zero_point = expected_level_low - fns.round(input_low / scale)
    zero_point = fns.clip(zero_point.astype(TensorDataType.int32), level_low, level_high)
    return scale, zero_point



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import Any, Dict, List, Optional, TypeVar

from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.utils.registry import Registry
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_dim import ElasticityDim
from nncf.torch.nncf_network import NNCFNetwork

ELASTICITY_HANDLERS_MAP = Dict[ElasticityDim, "ElasticityHandler"]
ElasticSearchSpace = TypeVar("ElasticSearchSpace")
ElasticityConfig = TypeVar("ElasticityConfig")
ELASTICITY_BUILDERS = Registry("Elasticity builder", add_name_as_attr=True)
ELASTICITY_PARAMS = Registry("Elasticity builder", add_name_as_attr=True)


class ElasticityHandler(ABC):
    """
    An interface for handling elasticity in the network. The elasticity defines variable values in properties of the
    layers or the network, e.g. variable number of channels in the Conv or variable number of layers in the network.
    By applying elasticity it's possible to derive a smaller models (Subnets) that have some elements in common with
    the original model.
    The interface defines methods for activation Subnets.
    """

    @abstractmethod
    def get_active_config(self) -> ElasticityConfig:
        """
        Forms an elasticity configuration that describes currently activated Subnet

        :return: elasticity configuration
        """

    @abstractmethod
    def get_random_config(self) -> ElasticityConfig:
        """
        Forms an elasticity configuration that describes a Subnet with randomly chosen elastic values

        :return: elasticity configuration
        """

    @abstractmethod
    def get_minimum_config(self) -> ElasticityConfig:
        """
        Forms an elasticity configuration that describes a Subnet with minimum elastic values

        :return: elasticity configuration
        """

    @abstractmethod
    def get_maximum_config(self) -> ElasticityConfig:
        """
        Forms an elasticity configuration that describes a Subnet with maximum elastic values

        :return: elasticity configuration
        """

    @abstractmethod
    def activate_supernet(self) -> None:
        """
        Activates the Supernet - the original network to which elasticity was applied.
        The Supernet is usually a maximum Subnet. But it might be not the case if maximum value of elastic properties
        is limited for efficiency of search. For instance, a supernet may contain 32 convolutional channels,
        but maximum subnet can be restricted to have 24 channels, at maximum.
        """

    @abstractmethod
    def activate_subnet_for_config(self, config: ElasticityConfig) -> None:
        """
        Activates a Subnet that corresponds to the given elasticity configuration

        :param config: elasticity configuration
        """

    @abstractmethod
    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """

    @abstractmethod
    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Initializes object from the state.

        :param state: Output of `get_state()` method.
        """

    def activate_random_subnet(self) -> None:
        """
        Activates a Subnet with random values of elastic properties.
        """
        config = self.get_random_config()
        self.activate_subnet_for_config(config)

    def activate_minimum_subnet(self) -> None:
        """
        Activates a minimum Subnet that corresponds to the minimum values of elastic properties.
        """
        config = self.get_minimum_config()
        self.activate_subnet_for_config(config)

    def activate_maximum_subnet(self) -> None:
        """
        Activates a maximum Subnet that corresponds to the maximum values of elastic properties.
        """
        config = self.get_maximum_config()
        self.activate_subnet_for_config(config)


class SEHandlerStateNames:
    ACTIVE_CONFIG = "active_config"


class SingleElasticityHandler(ElasticityHandler, ABC):
    """
    An interface for handling a single elasticity dimension in the network, e.g. elastic width or depth.
    """

    _state_names = SEHandlerStateNames

    @abstractmethod
    def get_search_space(self) -> ElasticSearchSpace:
        """
        :return: search space that is produced by iterating over all elastic parameters
        """

    @abstractmethod
    def get_transformation_commands(self) -> List[TransformationCommand]:
        """
        :return: transformation commands for introducing the elasticity to NNCFNetwork
        """

    @abstractmethod
    def resolve_conflicts_with_other_elasticities(
        self, config: ElasticityConfig, elasticity_handlers: ELASTICITY_HANDLERS_MAP
    ) -> ElasticityConfig:
        """
        Resolves a conflict between the given elasticity config and active elasticity configs of the given handlers.
        For example, elastic width configuration may contradict to elastic depth one. When we activate some
        configuration in the Elastic Width Handler, i.e. define number of output channels for some layers, we
        change output shapes of the layers. Consequently, it affects the blocks that can be skipped by Elastic Depth
        Handler, because input and output shapes may not be identical now.

        :param config: elasticity configuration
        :param elasticity_handlers: map of elasticity dimension to elasticity handler
        :return: elasticity configuration without conflicts with other active configs of other elasticity handlers
        """

    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Initializes object from the state.

        :param state: Output of `get_state()` method.
        """
        active_config = state[self._state_names.ACTIVE_CONFIG]
        self.activate_subnet_for_config(active_config)

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        active_config = self.get_active_config()
        return {
            self._state_names.ACTIVE_CONFIG: active_config,
        }


class BaseElasticityParams:
    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "BaseElasticityParams":
        """
        Creates the object from its config.
        """

    @classmethod
    @abstractmethod
    def from_state(cls, state: Dict[str, Any]) -> "BaseElasticityParams":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """

    @abstractmethod
    def get_state(self) -> Dict[str, Any]:
        """
        Returns the compression loss state.

        :return: The compression loss state.
        """


class SEHBuilderStateNames:
    ELASTICITY_PARAMS = "elasticity_params"


class SingleElasticityBuilder:
    """
    Determines which modifications should be made to the original FP32 model in order to introduce elasticity
    to the model.
    """

    _state_names = SEHBuilderStateNames

    def __init__(
        self,
        params: BaseElasticityParams,
        ignored_scopes: Optional[List[str]] = None,
        target_scopes: Optional[List[str]] = None,
    ):
        self._target_scopes = target_scopes
        self._ignored_scopes = ignored_scopes
        self._params = params

    @abstractmethod
    def build(self, target_model: NNCFNetwork) -> SingleElasticityHandler:
        """
        Creates modifications to the given NNCFNetwork for introducing elasticity and creates a handler object that
        can manipulate this elasticity.

        :param target_model: a target NNCFNetwork for adding modifications
        :return: a handler object that can manipulate the elasticity.
        """

    @abstractmethod
    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """

    @abstractmethod
    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Initializes object from the state.

        :param state: Output of `get_state()` method.
        """


def create_elasticity_builder_from_config(
    config: Dict[str, Any],
    elasticity_dim: ElasticityDim,
    ignored_scopes: Optional[List[str]] = None,
    target_scopes: Optional[List[str]] = None,
) -> SingleElasticityBuilder:
    params_cls = ELASTICITY_PARAMS.get(elasticity_dim)
    params = params_cls.from_config(config)
    elasticity_builder_cls = ELASTICITY_BUILDERS.get(elasticity_dim)
    return elasticity_builder_cls(params, ignored_scopes, target_scopes)
